from __future__ import annotations

import itertools
import os
import time
from contextlib import ExitStack, contextmanager
from copy import deepcopy
from dataclasses import dataclass, field
from datetime import datetime
from io import BytesIO, IOBase
from json import loads as json_loads
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import httpx
import structlog
import tenacity
from httpx import Request, Response
from pydantic import BaseModel

from avatars.arrow_utils import (
    ArrowStreamReader,
    ArrowStreamWriter,
    FileLike,
    FileLikes,
    is_text_file_or_buffer,
)
from avatars.models import JobStatus
from avatars.utils import ContentType, ensure_valid, pop_or, remove_optionals, validated

logger = structlog.getLogger(__name__)
structlog.configure(logger_factory=structlog.stdlib.LoggerFactory())

DEFAULT_RETRY_TIMEOUT = 60
DEFAULT_RETRY_INTERVAL = 5
DEFAULT_RETRY_COUNT = 20
DEFAULT_TIMEOUT = 60 * 4
DEFAULT_PER_CALL_TIMEOUT = 15

IN_PROGRESS_STATUSES = (JobStatus.pending, JobStatus.started)

DEFAULT_BINARY_CONTENT_TYPES = (
    ContentType.PDF,
    ContentType.OCTET_STREAM,
    ContentType.ARROW_STREAM,
)
DEFAULT_TEXT_CONTENT_TYPES = (ContentType.CSV, ContentType.JSON)

T = TypeVar("T")
R = TypeVar("R")
RequestClass = TypeVar("RequestClass", bound=BaseModel)
ResponseClass = TypeVar("ResponseClass", bound=BaseModel)

JsonLike = dict[str, Any]

Content = Union[Iterable[bytes], bytes]
ContentBuilderFunc = Optional[Callable[..., Content]]
UserContent = Union[JsonLike, str, bytes, Optional[BytesIO]]
StreamedContent = Optional[Union[BytesIO, bytes, str]]

AuthRefreshFunc = Optional[Callable[..., dict[str, str]]]


def _get_nested_value(
    obj: Union[Mapping[Any, Any], Sequence[Any]], key: str, default: Any = None
) -> Any:
    """
    Return value from (possibly) nested key in JSON dictionary.
    """
    if isinstance(obj, Sequence) and not isinstance(obj, str):
        for item in obj:
            return _get_nested_value(item, key, default=default)

    if isinstance(obj, Mapping):
        if key in obj:
            return obj[key]
        return _get_nested_value(list(obj.values()), key, default=default)

    return default


def _log_before_retry_attempt(retry_state: tenacity.RetryCallState) -> None:
    if not retry_state.next_action:
        logger.info(" log_before_retry_attempt no next action")
        return
    next_retry_seconds = retry_state.next_action.sleep
    logger.info(f"Retrying in {next_retry_seconds}s...")


def _log_after_failure(
    retry_state: tenacity.RetryCallState, *, data: ContextData
) -> None:
    if not retry_state.outcome:
        return

    if retry_state.outcome.failed:
        e = retry_state.outcome.exception()
        error_message = str(e)
        logger.warning(error_message, url=data.url, base_url=data.base_url)


def _reraise_on_timeout(
    retry_state: tenacity.RetryCallState, *, data: ContextData
) -> None:
    """After last retry attempt, if the outcome is a timeout, raise a custom exception."""
    if not retry_state.outcome:
        return

    if retry_state.outcome.failed:
        e = retry_state.outcome.exception()
        if isinstance(e, (httpx.ReadTimeout, httpx.WriteTimeout)):
            msg = f"Timeout waiting for {data.method.upper()} on {data.url}"
            raise Timeout(msg)
        raise e  # type: ignore[misc] # Exception must be derived from BaseException


class Timeout(Exception):
    pass


class ClientTimeout:
    def __init__(self, *, timeout: float, max_seconds: float = 10):
        self.retry_seconds: float = timeout
        self.max_seconds: float = max_seconds
        self.is_active: bool = False
        self.last_time: float = 0
        self.elapsed_seconds: float = 0
        self.interval: Optional[Iterator[float]] = None

    def start(self) -> None:
        self.is_active = True
        self.last_time = time.time()
        self.elapsed_seconds = 0

        # Exponential interval, capped at max_interval
        self.sleep_interval = iter(
            min(2**i, self.max_seconds) for i in itertools.count()
        )

    def stop(self) -> None:
        self.is_active = False

    def update(self) -> None:
        now = time.time()
        self.elapsed_seconds += now - self.last_time
        self.last_time = now

    def is_expired(self) -> bool:
        return self.elapsed_seconds > self.retry_seconds

    def next_interval(self) -> float:
        interval = next(self.sleep_interval)

        if self.elapsed_seconds > self.retry_seconds:
            interval = min(interval, self.retry_seconds - self.elapsed_seconds)

        return float(interval)

    def intervals(self, label: str) -> Iterator[float]:
        self.start()

        # Iterate while we are < retry_timeout
        while not self.is_expired() and self.is_active:
            yield self.next_interval()
            self.update()

        if self.is_active and self.is_expired():
            raise Timeout(f"{label} expired")


@dataclass
class ContextData:
    base_url: str
    method: str
    url: str
    headers: dict[str, str]
    http_request: Optional[Request] = None
    http_response: Optional[Response] = None
    timeout: float = 0.0
    params: Optional[Dict[str, Any]] = None
    json_data: Optional[BaseModel] = None
    form_data: Optional[Union[BaseModel, Dict[str, Any]]] = None
    files: Optional[FileLikes] = None
    content_builder: ContentBuilderFunc = None
    should_verify_auth: bool = True
    should_stream: bool = False
    destination: Optional[FileLike] = None
    want_content: bool = False

    def update(self, **kwargs: Any) -> None:
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)

    def build_params_arg(self) -> Optional[Dict[str, Any]]:
        return remove_optionals(self.params)

    def build_json_data_arg(self) -> Optional[str]:
        return json_loads(self.json_data.model_dump_json()) if self.json_data else None

    def build_form_data_arg(self) -> Optional[Dict[str, Any]]:
        arg = (
            self.form_data.model_dump()
            if isinstance(self.form_data, BaseModel)
            else self.form_data
        )

        return remove_optionals(arg)

    def build_files_arg(self) -> Optional[list[Tuple[str, Any]]]:
        return [("file", file) for file in self.files] if self.files else None

    def build_content_arg(self) -> Optional[Content]:
        return self.content_builder() if self.content_builder else None

    def status_is(self, status_code: int) -> bool:
        if self.http_response:
            return self.http_response.status_code == status_code

        return False

    def ensure_status_is(self, status_code: int) -> bool:
        if self.http_response:
            if not self.status_is(status_code):
                raise Exception(
                    f"Expected code {status_code}"
                    f" for {self.method} on {self.url}"
                    f", got {self.http_response.status_code} instead"
                )

            return True
        else:
            raise Exception("Expected a HTTP response, got None instead")

    def has_header(self, header: str) -> bool:
        if self.http_response and header in self.http_response.headers:
            return True

        return False

    def get_header(self, header: str) -> Any:
        return ensure_valid(self.http_response).headers[header]

    def update_headers(self, headers: dict[str, str]) -> None:
        self.headers |= headers

        if self.http_request:
            self.http_request.headers.update(headers)

    def content_type(self) -> ContentType:
        return ContentType(self.get_header("content-type").split(";")[0].strip())

    def is_created(self) -> bool:
        return self.status_is(httpx.codes.CREATED) and self.has_header("location")

    def ensure_created(self) -> bool:
        return self.ensure_status_is(httpx.codes.CREATED) and self.has_header(
            "location"
        )

    def is_content_json(self) -> bool:
        return bool(self.content_type() == ContentType.JSON)

    def is_content_arrow(self) -> bool:
        return bool(self.content_type() == ContentType.ARROW_STREAM)

    def is_content_text(self) -> bool:
        return self.content_type() in DEFAULT_TEXT_CONTENT_TYPES

    def is_content_binary(self) -> bool:
        return self.content_type() in DEFAULT_BINARY_CONTENT_TYPES

    def get_user_content(self) -> UserContent:
        with validated(self.http_response, "response") as resp:
            if self.should_stream:
                return self.stream_response()
            else:
                if self.is_content_json():
                    return self.response_to_json()
                elif self.is_content_binary():
                    return resp.content
                else:
                    return resp.text

    def get_content_message(self) -> str:
        content = self.get_user_content()
        msg = ""

        if isinstance(content, dict):
            if "detail" in content:
                if "message" in content["detail"]:
                    msg = content["detail"]["message"]
        elif isinstance(content, str):
            msg = content
        else:
            logger.warning(f"Expected readable content, got {type(content)} instead")

        return msg

    def response_to_json(self) -> dict[str, Any]:
        resp = ensure_valid(self.http_response, "response")

        as_json: Dict[str, Any] = {}

        if self.is_content_json():
            as_json = resp.json()

        return as_json

    def stream_response_content(self, destination: FileLike) -> None:
        with validated(self.http_response, "response") as resp:
            if self.is_content_arrow():
                with ArrowStreamReader() as reader:
                    reader.write_parquet(destination, resp.iter_bytes())
            else:
                try:
                    if is_text_file_or_buffer(destination):
                        for chunk in resp.iter_text():
                            destination.write(chunk)  # type: ignore[call-overload]
                    else:
                        # Assume bytes...
                        for chunk in resp.iter_bytes():  # type: ignore[assignment]
                            destination.write(chunk)  # type: ignore[call-overload]
                finally:
                    resp.close()

    def stream_response(
        self, destination: Optional[FileLike] = None
    ) -> StreamedContent:
        """
        Handle the streaming of a response to a destination.

        If the destination is not provided, it returns the content as bytes.

        This needs the httpx.Client instance to remain open, even though no client
        is used in this function.

        Parameters
        ----------
        response
            The response object to be streamed.
        destination
            The destination where the response will be streamed.
            If not provided, the content is returned as.

        Returns
        -------
            If no destination was provided, it returns the raw bytes.
        """
        content: StreamedContent = None
        buffer = BytesIO()
        opened = False

        if isinstance(destination, str):
            opened = True
            destination_data = open(destination, "wb")
        else:
            destination_data = destination or buffer

        self.stream_response_content(destination_data)

        if isinstance(destination_data, IOBase):
            logger.info(f"base_client: flushing {destination=}")
            destination_data.flush()

        if opened:
            logger.info(f"base_client: closing {destination=}")
            destination_data.close()

        buffer.seek(0, os.SEEK_SET)

        content = buffer if not destination else None

        if self.want_content and content:
            content = content.read()

            if self.is_content_text():
                content = content.decode()

        return content

    def clone(self) -> ContextData:
        return deepcopy(self)


@dataclass
class OperationInfo:
    data: ContextData
    in_progress: bool = False
    last_updated_at: datetime = field(default_factory=datetime.now)
    response: Optional[Any] = None


class ClientContext:
    def __init__(
        self,
        http_client: httpx.Client,
        data: ContextData,
        on_auth_refresh: AuthRefreshFunc = None,
    ) -> None:
        self.http_client: httpx.Client = http_client
        self.data: ContextData = data
        self.timeout: Optional[ClientTimeout] = None
        self.on_auth_refresh = on_auth_refresh

    def build_request(self) -> Request:
        self.data.http_request = self.http_client.build_request(
            method=self.data.method,
            url=self.data.url,
            params=self.data.build_params_arg(),
            json=self.data.build_json_data_arg(),
            data=self.data.build_form_data_arg(),
            files=self.data.build_files_arg(),
            content=self.data.build_content_arg(),
            headers=self.data.headers,
            timeout=None,
        )

        return ensure_valid(self.data.http_request)

    def retry(
        self,
        retry_count: int = DEFAULT_RETRY_COUNT + 1,  # stop is inclusive
        retry_inverval: int = DEFAULT_RETRY_INTERVAL,
    ) -> Iterator[tenacity.AttemptManager]:
        for attempt in tenacity.Retrying(
            stop=tenacity.stop_after_attempt(retry_count),
            wait=tenacity.wait_exponential(max=retry_inverval),
            before_sleep=_log_before_retry_attempt,
            retry_error_callback=lambda call_state: _reraise_on_timeout(
                call_state, data=self.data
            ),
            after=lambda call_state: _log_after_failure(call_state, data=self.data),
            reraise=True,
        ):
            yield attempt

    def send_request(self) -> Response:
        first_or_retry_all = True

        while first_or_retry_all:
            first_or_retry_all = False
            request = ensure_valid(self.data.http_request)

            for attempt in self.retry(DEFAULT_RETRY_COUNT + 1, DEFAULT_RETRY_INTERVAL):
                with attempt:
                    self.data.http_response = self.http_client.send(
                        request=request,
                        stream=self.data.should_stream,
                    )

                    if self.check_auth_refreshed():
                        # Reset/rebuild current request
                        first_or_retry_all = True
                        self.build_request()
                        break

        self.check_success()

        return ensure_valid(self.data.http_response)

    def send_request_and_build_response(
        self, response_cls: type[ResponseClass]
    ) -> ResponseClass:
        self.send_request()

        return self.build_response(response_cls)

    def build_and_send_request(self) -> Any:
        self.build_request()

        return self.send_request()

    def build_response(self, response_cls: type[ResponseClass]) -> ResponseClass:
        return response_cls(**self.data.response_to_json())

    def cancel_timeout(self) -> None:
        if self.timeout:
            self.timeout.stop()

    def wait_intervals(self, label: str) -> Iterator[float]:
        self.timeout = ClientTimeout(timeout=DEFAULT_RETRY_TIMEOUT)
        self.timeout.start()

        for interval in self.timeout.intervals(label):
            yield interval

        self.timeout = None

    def loop_until(
        self,
        *,
        label: str,
        update_func: Callable[..., bool],
        response_cls: Optional[Type[ResponseClass]] = None,
    ) -> OperationInfo:
        def call_update_func(info: OperationInfo) -> bool:
            if response_cls:
                info.response = self.build_response(response_cls)
                return update_func(info, info.response)
            else:
                return update_func(info)

        info = OperationInfo(data=self.data)
        what = str(response_cls) if response_cls else "request"
        what_label = f"for {what} at {self.data.url} to complete"
        loops = 1

        self.build_request()
        info.in_progress = True

        while info.in_progress:
            self.send_request()

            stop = call_update_func(info)

            if stop or not info.in_progress:
                break

            logger.info(
                f"waiting {what_label}"
                f"(loop {loops}, sleeping {DEFAULT_RETRY_INTERVAL}s)"
            )
            time.sleep(DEFAULT_RETRY_INTERVAL)

            loops += 1

        if not response_cls:
            info.response = self.data.get_user_content()

        return info

    def check_success(self) -> None:
        resp = ensure_valid(self.data.http_response, "response")

        if resp.is_success:
            return

        self.raise_on_status(resp)

    def check_auth_refreshed(
        self,
    ) -> bool:
        refreshed = False

        if self.data.status_is(httpx.codes.UNAUTHORIZED):
            msg = self.data.get_content_message()

            if "credentials expired" in msg:
                if self.on_auth_refresh:
                    logger.info("trying to refresh authentication token")
                    new_headers = self.on_auth_refresh()
                    self.data.update_headers(new_headers)
                    logger.info("authentication refreshed, retrying previous request")
                    refreshed = True
                else:
                    logger.warning("Authentication refresh needed but not configured")

        return refreshed

    def check_authenticated(self, resp: Response, content: dict[str, Any]) -> None:
        value = content.get("detail")

        if (
            resp.status_code == httpx.codes.UNAUTHORIZED
            and isinstance(value, str)
            and "authenticated" in value
        ):
            raise Exception("You are not authenticated.")

    def raise_on_status(self, resp: Response) -> None:
        content = self.data.get_user_content() or "no message available"

        if isinstance(content, dict):
            self.check_authenticated(resp, content)

            standard_error = _get_nested_value(content, "message")

            if standard_error:
                error_msg = standard_error
            elif validation_error := _get_nested_value(content, "msg"):
                if detailed_message := _get_nested_value(content, "loc"):
                    field = detailed_message[-1]
                    error_msg = f"{validation_error}: {field}"
                else:
                    error_msg = f"Bad Request: {validation_error}"
            else:
                error_msg = "Internal error: " + str(content)
        else:
            error_msg = content

        raise Exception(
            f"Got error in HTTP request: {self.data.method} {self.data.url}."
            f" Error status {resp.status_code} - {error_msg}"
        )


def update_request_op(info: OperationInfo) -> bool:
    info.in_progress = info.data.status_is(httpx.codes.ACCEPTED)
    return False


def update_response_op(info: OperationInfo, response: ResponseClass) -> bool:
    ret: bool = False

    if hasattr(response, "status"):
        info.in_progress = response.status in IN_PROGRESS_STATUSES
    else:
        ret = update_request_op(info)

    if hasattr(response, "last_updated_at"):
        info.last_updated_at = response.last_updated_at

    return ret


class BaseClient:
    def __init__(
        self,
        base_url: str,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
        should_verify_ssl: bool = True,
        *,
        verify_auth: bool = True,
        on_auth_refresh: Optional[AuthRefreshFunc] = None,
        http_client: Optional[httpx.Client] = None,
        headers: Dict[str, str] = {},
    ) -> None:
        """Client to communicate with the Avatar API.

        Parameters
        ----------
        base_url
            url of the API
        timeout:
            timeout in seconds, by default DEFAULT_TIMEOUT
        should_verify_ssl :, optional
            whether to verify SSL certificates on the server. By default True
        http_client :, optional
            allow passing in custom httpx.Client instance, by default None
        verify_auth :, optional
            Bypass client-side authentication verification, by default True
        """
        if '"' in base_url:
            raise ValueError(
                f"Expected base_url not to contain quotes. Got {base_url} instead"
            )

        self.base_url = base_url
        self.timeout = timeout
        self.should_verify_ssl = should_verify_ssl
        self.verify_auth = verify_auth
        self._on_auth_refresh = on_auth_refresh
        self._http_client = http_client
        self._headers = {"Avatars-Accept-Created": "yes"} | headers

    def set_header(self, key: str, value: str) -> None:
        self._headers[key] = value

    def on_auth_refresh(
        self, on_auth_refresh: Optional[AuthRefreshFunc] = None
    ) -> None:
        self._on_auth_refresh = on_auth_refresh

    def prepare_files(
        self, stack: ExitStack, headers: dict[str, Any], keyword_args: dict[str, Any]
    ) -> Optional[FileLikes]:
        files: Any = pop_or(keyword_args, "files", [])
        files = files if isinstance(files, list) else [files]

        if f := pop_or(keyword_args, "file", None):
            files.append(f)

        prepared_files: Optional[FileLikes] = None

        if files:
            prepared_files = []

            for f in files:
                if isinstance(f, (str, Path)) and Path(f).is_file():
                    prepared_files.append(stack.enter_context(open(f, "rb")))
                else:
                    raise ValueError(
                        f"Expected streamable file-like object, got {f} instead"
                    )

        return prepared_files

    def make_content_builder(
        self, headers: dict[str, Any], keyword_args: dict[str, Any]
    ) -> ContentBuilderFunc:
        content_builder: ContentBuilderFunc = None

        if "content" in keyword_args:
            content = keyword_args.pop("content")
            content_builder = lambda: content  # noqa
        elif "dataset" in keyword_args:
            ds = keyword_args.pop("dataset")
            headers["Content-Type"] = ContentType.ARROW_STREAM.value
            content_builder = lambda: ArrowStreamWriter(ds)  # noqa

        return content_builder

    @contextmanager
    def context(
        self, *, ctx: Optional[ClientContext] = None, **kwargs: Any
    ) -> Generator[ClientContext, None, None]:
        with ExitStack() as stack:
            http_client = self._http_client or httpx.Client(
                base_url=self.base_url,
                timeout=self.timeout,
                verify=self.should_verify_ssl,
            )

            # Grab special keys
            headers: dict[str, Any] = pop_or(kwargs, "headers", {})
            files = self.prepare_files(stack, headers, kwargs)
            content_builder = self.make_content_builder(headers, kwargs)
            want_content: bool = pop_or(kwargs, "want_content", False)

            if not ctx:
                ctx = ClientContext(
                    http_client=http_client,
                    data=ContextData(
                        base_url=self.base_url, headers=self._headers.copy(), **kwargs
                    ),
                    on_auth_refresh=self._on_auth_refresh,
                )

            ctx.data.update(**kwargs)
            ctx.data.headers.update(headers)
            ctx.data.files = files
            ctx.data.content_builder = content_builder
            ctx.data.want_content = want_content

            yield ctx

            ctx.data.files = None
            ctx.data.content_builder = None
            ctx.data.want_content = False

    def create(
        self,
        *,
        url: str,
        request: RequestClass,
        response_cls: Type[ResponseClass],
        timeout: Optional[int] = None,
    ) -> ResponseClass:
        with self.context(method="post", url=url) as ctx:
            ctx.data.json_data = request
            ctx.build_and_send_request()
            ctx.data.ensure_created()

            info = self.wait_created(
                url=ctx.data.get_header("location"),
                update_func=update_response_op,
                response_cls=response_cls,
                ctx=ctx,
            )

            return cast(ResponseClass, info.response)

    def request(self, method: str, url: str, **kwargs: Any) -> Any:
        response: Any = None

        with self.context(method=method, url=url, **kwargs) as ctx:
            ctx.build_and_send_request()
            destination = kwargs.get("destination", None)

            if destination:
                response = ctx.data.stream_response(destination=destination)
            elif ctx.data.is_created():
                response = self.wait_created(
                    url=ctx.data.get_header("location"),
                    update_func=update_request_op,
                    ctx=ctx,
                ).response
            else:
                response = ctx.data.get_user_content()

        return response

    def wait_created(
        self, *, ctx: ClientContext, url: str, **kwargs: Any
    ) -> OperationInfo:
        with self.context(method="get", url=url, ctx=ctx) as ctx:
            return ctx.loop_until(
                label=f"created resource is ready at {url}",
                **kwargs,
            )

    def check_auth(self, data: ContextData) -> None:
        should_verify = self.verify_auth and data.should_verify_auth

        if should_verify and "Authorization" not in self._headers:
            raise Exception("You are not authenticated.")

    def send_request(self, *, method: str, url: str, **kwargs: Any) -> Any:
        """Request the API."""
        with self.context(method=method, url=url, **kwargs) as ctx:
            self.check_auth(ctx.data)
            ctx.build_and_send_request()

            return ctx.data.get_user_content()

    def get_job(
        self,
        *,
        url: str,
        response_cls: Type[ResponseClass],
        **kwargs: Any,
    ) -> ResponseClass:
        with self.context(method="get", url=url) as ctx:
            ctx.build_and_send_request()

            info = ctx.loop_until(
                label=f"get job at {url}",
                update_func=update_response_op,
                response_cls=response_cls,
            )

            return cast(ResponseClass, info.response)

    def __str__(self) -> str:
        return ", ".join(
            f"ApiClient(base_url={self.base_url}"
            f"timeout={self.timeout}"
            f"should_verify_ssl={self.should_verify_ssl}"
            f"verify_auth={self.verify_auth})"
        )



import pandas as pd
import pytest

from avatars.lib.split import (
    get_index_without_duplicated_modalities,
    get_split_for_batch,
)


def test_get_split_for_batch() -> None:
    df = pd.DataFrame(
        data={
            "a": [1, 3, 5, 1, 3, 5, 1, 3, 5, 1, 3, 5],
            "b": ["a", "b", "a", "a", "b", "a", "a", "b", "a", "a", "b", "a"],
        }
    )
    result = get_split_for_batch(df, row_limit=6, seed=42)

    training = pd.DataFrame(
        data={"a": [1, 3, 3, 1, 3, 5], "b": ["a", "b", "b", "a", "b", "a"]},
        index=[0, 1, 10, 3, 7, 2],
    )
    batch = pd.DataFrame(
        data={"a": [1, 3, 5, 1, 5, 5], "b": ["a", "b", "a", "a", "a", "a"]},
        index=[9, 4, 11, 6, 5, 8],
    )

    pd.testing.assert_frame_equal(result[0], training)
    pd.testing.assert_frame_equal(result[1][0], batch)


def test_get_split_for_batch_with_optimization() -> None:
    # the four first record are unique but only the two first should be kept by the subset process
    df = pd.DataFrame(
        data={
            "a": ["a", "b", "b", "a", "a", "b", "b", "a", "a", "b", "b", "a"],
            "b": ["A", "B", "B", "B", "A", "B", "B", "B", "A", "B", "B", "B"],
            "c": ["1", "2", "1", "1", "1", "2", "1", "1", "1", "2", "1", "1"],
        }
    )
    training, splits = get_split_for_batch(df, row_limit=6, seed=42)

    # the two first record should be the same of the original data and the other ones
    # should be randomly selected
    expected_training = pd.DataFrame(
        data={
            "a": ["a", "b", "b", "a", "a", "b"],
            "b": ["A", "B", "B", "B", "B", "B"],
            "c": ["1", "2", "1", "1", "1", "1"],
        },
        index=[0, 1, 10, 3, 7, 2],
    )
    expected_batch = pd.DataFrame(
        data={
            "a": ["b", "a", "a", "b", "b", "a"],
            "b": ["B", "A", "B", "B", "B", "A"],
            "c": ["2", "1", "1", "1", "2", "1"],
        },
        index=[9, 4, 11, 6, 5, 8],
    )
    pd.testing.assert_frame_equal(expected_training, training)
    pd.testing.assert_frame_equal(expected_batch, splits[0])


def test_get_split_for_batch_raise_error() -> None:
    df = pd.DataFrame(data={"a": [1, 3], "b": ["a", "b"]})

    with pytest.raises(
        ValueError, match="You have too much variability in your categorical data"
    ):
        get_split_for_batch(df, row_limit=2, seed=42)


def test_get_index_without_duplicated_modalities() -> None:
    df = pd.DataFrame({"a": [1, 2, 1, 3, 2, 1], "b": [1, 2, 3, 3, 2, 1]})

    expect = [0, 1, 2, 3]
    result = get_index_without_duplicated_modalities(df)
    assert expect, result



from datetime import datetime

import numpy as np
import pandas as pd
import pytest

from avatars.lib.split_columns_types import split_columns_types


@pytest.fixture
def many_dtypes_df() -> pd.DataFrame:
    df = pd.DataFrame()
    df["ints"] = [-2, 0, 2, 3, 4, 5, 6]
    df["floats"] = [-2.1, 0.0, 2.1, 3.1, 4.1, 5.1, 6.1]
    df["strings"] = ["Mr", "John", "Doe", "Mrs.", "Mary", "Higgins", "Clark"]
    df["datetimes"] = pd.to_datetime([datetime(2022, 1, 1).replace(tzinfo=None)] * 7)
    return df


def test_split_column_types(many_dtypes_df: pd.DataFrame) -> None:
    test_categorical = [2, 3]
    test_continuous = [0, 1]

    continuous, categorical = split_columns_types(many_dtypes_df)

    assert categorical == test_categorical
    assert continuous == test_continuous


def test_split_column_types_no_categorical(many_dtypes_df: pd.DataFrame) -> None:
    test_continuous, test_categorical = split_columns_types(
        many_dtypes_df[["ints", "floats"]]
    )

    assert len(test_categorical) == 0
    assert test_continuous == [0, 1]


def test_split_column_types_no_continuous(many_dtypes_df: pd.DataFrame) -> None:
    test_continuous, test_categorical = split_columns_types(
        many_dtypes_df[["strings", "datetimes"]]
    )

    assert test_categorical == [0, 1]
    assert len(test_continuous) == 0


@pytest.mark.parametrize("dtype_int", [np.int8, np.int16, np.int32, np.int64])
def test_split_column_types_different_int(
    many_dtypes_df: pd.DataFrame, dtype_int: np.int_
) -> None:
    test_categorical = [2, 3]
    test_continuous = [0, 1]

    df = many_dtypes_df
    df["ints"] = df["ints"].astype(dtype_int)
    continuous, categorical = split_columns_types(df)

    assert categorical == test_categorical
    assert continuous == test_continuous


@pytest.mark.parametrize("dtype_float", [np.float64, np.float32, np.float16])
def test_split_column_types_different_float(
    many_dtypes_df: pd.DataFrame, dtype_float: np.float_
) -> None:
    test_categorical = [2, 3]
    test_continuous = [0, 1]

    df = many_dtypes_df
    df["floats"] = df["floats"].astype(dtype_float)
    continuous, categorical = split_columns_types(df)

    assert categorical == test_categorical
    assert continuous == test_continuous



# This file has been generated - DO NOT MODIFY
# API Version : 10.2.0

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional, Protocol, Union, runtime_checkable
from uuid import UUID

import pandas as pd
from pydantic import BaseModel, ConfigDict, Field, RootModel
from typing_extensions import Annotated

# generated by datamodel-codegen:
#   filename:  input.json


class AdviceParameters(BaseModel):
    """
    Parameters for advice job.
    """

    model_config = ConfigDict(
        extra="allow",
    )
    dataset_id: Annotated[UUID, Field(title="Dataset Id")]


class AnalysisStatus(Enum):
    started = "started"
    done = "done"


class ClosestRatePercentageThresholdItem(RootModel[float]):
    root: Annotated[float, Field(ge=0.0, le=100.0)]


class ClosestRateRatioThresholdItem(RootModel[float]):
    root: Annotated[float, Field(ge=0.0, le=1.0)]


class ClusterStats(BaseModel):
    n_active_tasks: Annotated[Optional[int], Field(title="N Active Tasks")] = None
    available_concurrency: Annotated[
        Optional[int], Field(title="Available Concurrency")
    ] = None
    utilization_rate_100: Annotated[
        Optional[int], Field(title="Utilization Rate 100")
    ] = None
    status_message: Annotated[Optional[str], Field(title="Status Message")] = ""


class ColumnStats(BaseModel):
    label: Annotated[str, Field(title="Label")]
    dtype: Annotated[str, Field(title="Dtype")]
    missing_number: Annotated[int, Field(ge=0, title="Missing Number")]
    missing_proportion: Annotated[
        float, Field(ge=0.0, le=1.0, title="Missing Proportion")
    ]
    unique_number: Annotated[int, Field(ge=0, title="Unique Number")]
    unique_proportion: Annotated[
        float, Field(ge=0.0, le=1.0, title="Unique Proportion")
    ]
    mean: Annotated[Optional[float], Field(title="Mean")] = None
    median: Annotated[Optional[float], Field(title="Median")] = None
    standard_deviation: Annotated[
        Optional[float], Field(title="Standard Deviation")
    ] = None
    mode: Annotated[Optional[Union[str, int, float]], Field(title="Mode")] = None
    is_skewed: Annotated[Optional[bool], Field(title="Is Skewed")] = None
    is_rare: Annotated[Optional[bool], Field(title="Is Rare")] = None


class ColumnType(Enum):
    int = "int"
    bool = "bool"
    category = "category"
    float = "float"
    datetime = "datetime"


class CompatibilityStatus(Enum):
    compatible = "compatible"
    incompatible = "incompatible"
    unknown = "unknown"


class Contributions(BaseModel):
    data: Annotated[Dict[str, Dict[str, float]], Field(title="Data")]


class DatasetSummary(BaseModel):
    stats: Annotated[List[ColumnStats], Field(title="Stats")]
    probable_ids: Annotated[List[str], Field(title="Probable Ids")]


class ExcludeVariablesMethod(Enum):
    """
    The method to exclude column.

    There are several possible choices:

    - ``row_order``: **SENSITIVE** The excluded column will be linked to the original row order.
    This is a violation of privacy.

    - ``coordinate_similarity``: The excluded column will be linked by individual similarity.
    """

    row_order = "row_order"
    coordinate_similarity = "coordinate_similarity"


class ExcludeVariablesParameters(BaseModel):
    """
    Parameters to exclude some variables from the anonymization before re-assigning.

    This parameter can be applied to any variable, regardless of its type but it is particularly
    recommended when the data contains categorical variables with a large number of modalities.
    """

    variable_names: Annotated[
        List[str],
        Field(
            description="List of variables to exclude from the anonymization. ",
            title="Variable Names",
        ),
    ]
    replacement_strategy: Annotated[
        ExcludeVariablesMethod,
        Field(description="See :class:`ExcludeVariablesMethod`."),
    ]


class ExplainedVariance(BaseModel):
    raw: Annotated[List[float], Field(title="Raw")]
    ratio: Annotated[List[float], Field(title="Ratio")]


class ProjectionType(Enum):
    fpca = "fpca"


class FPCAParameters(BaseModel):
    """
    Parameters of a FPCA projection.
    """

    nf: Annotated[
        int,
        Field(
            description="Number of degrees of freedom kept by the projection.",
            title="Nf",
        ),
    ]
    projection_type: Annotated[ProjectionType, Field(title="Projection Type")] = (
        ProjectionType.fpca
    )


class FileType(Enum):
    csv = "csv"
    parquet = "parquet"


class ProjectionType1(Enum):
    flatten = "flatten"


class FlattenParameters(BaseModel):
    """
    Parameters of a Flatten projection.
    """

    projection_type: Annotated[ProjectionType1, Field(title="Projection Type")] = (
        ProjectionType1.flatten
    )


class ForgottenPasswordRequest(BaseModel):
    email: Annotated[str, Field(title="Email")]


class GenericParameters(BaseModel):
    pass
    model_config = ConfigDict(
        extra="allow",
    )


class GenericResult(BaseModel):
    pass


class ProjectionType2(Enum):
    geolocation_density = "geolocation_density"


class GeolocationDensityParameters(BaseModel):
    """
    Parameters of a Geolocation Density projection.
    """

    projection_type: Annotated[ProjectionType2, Field(title="Projection Type")] = (
        ProjectionType2.geolocation_density
    )
    n_lon_to_compute: Annotated[
        Optional[int],
        Field(description="Number of longitude to compute.", title="N Lon To Compute"),
    ] = 100
    n_lat_to_compute: Annotated[
        Optional[int],
        Field(description="Number of latitude to compute.", title="N Lat To Compute"),
    ] = 100


class GeolocationFeatures(Enum):
    """
    Features available for geolocation feature projection and metrics.
    """

    length = "length"
    duration = "duration"
    speed = "speed"


class ProjectionType3(Enum):
    geolocation_features = "geolocation_features"


class GeolocationFeaturesParameters(BaseModel):
    """
    Parameters of a Geolocation Features projection.
    """

    projection_type: Annotated[ProjectionType3, Field(title="Projection Type")] = (
        ProjectionType3.geolocation_features
    )


class GeolocationPrivacyMetrics(BaseModel):
    local_cloaking: Annotated[float, Field(title="Local Cloaking")]
    hidden_rate: Annotated[float, Field(title="Hidden Rate")]
    specific_hidden_rate: Annotated[
        Optional[float], Field(title="Specific Hidden Rate")
    ] = None
    distance_to_closest: Annotated[
        Optional[float], Field(title="Distance To Closest")
    ] = None
    closest_distances_ratio: Annotated[
        Optional[float], Field(title="Closest Distances Ratio")
    ] = None
    closest_rate: Annotated[Optional[float], Field(title="Closest Rate")] = None
    features_inference_rate: Annotated[
        Optional[float], Field(title="Features Inference Rate")
    ] = None
    features_correlation_protection_rate: Annotated[
        Optional[float], Field(title="Features Correlation Protection Rate")
    ] = None
    thresholded_signal_inference_rate: Annotated[
        Optional[float], Field(title="Thresholded Signal Inference Rate")
    ] = None
    trace_correlation_protection_rate: Annotated[
        Optional[float], Field(title="Trace Correlation Protection Rate")
    ] = None
    inference_distances: Annotated[
        Optional[List[float]], Field(title="Inference Distances")
    ] = None


class TrainingFractionItem(RootModel[float]):
    root: Annotated[float, Field(ge=0.0, le=1.0)]


class ImputeMethod(Enum):
    knn = "knn"
    mode = "mode"
    median = "median"
    mean = "mean"
    fast_knn = "fast_knn"


class JobKind(Enum):
    avatarization = "avatarization"
    privacy_metrics = "privacy_metrics"
    signal_metrics = "signal_metrics"
    avatarization_with_time_series = "avatarization_with_time_series"
    avatarization_multi_table = "avatarization_multi_table"
    privacy_metrics_with_time_series = "privacy_metrics_with_time_series"
    privacy_metrics_multi_table = "privacy_metrics_multi_table"
    signal_metrics_with_time_series = "signal_metrics_with_time_series"
    privacy_metrics_geolocation = "privacy_metrics_geolocation"
    generic = "generic"
    advice = "advice"


class JobProgress(BaseModel):
    completion_rate_100: Annotated[int, Field(title="Completion Rate 100")]
    name: Annotated[Optional[str], Field(title="Name")] = None
    created_at: Annotated[datetime, Field(title="Created At")]


class JobStatus(Enum):
    pending = "pending"
    started = "started"
    success = "success"
    failure = "failure"
    killed = "killed"
    unknown = "unknown"


class LinkMethod(Enum):
    """
    Available assignment methods to link a child to its parent table after the anonymization.

    LINEAR_SUM_ASSIGNMENT: Assign using the linear sum assignment algorithm.
    This method is a good privacy and utility trade-off. The algorithm consumes lots of resources.

    MINIMUM_DISTANCE_ASSIGNMENT: Assign using the minimum distance assignment algorithm.
    This method assigns the closest child to the parent. It is an acceptable privacy and utility
    trade-off.
    This algorithm consumes less resources than the linear sum assignment.

    SENSITIVE_ORIGINAL_ORDER_ASSIGNMENT: Assign the child to the parent using the original order.
    WARNING!!! This method is a HIGH PRIVACY BREACH as it keeps the original order to assign
    the child to the parent.
    This method isn't recommended for privacy reasons but consumes less resources than the other
    methods.
    """

    linear_sum_assignment = "linear_sum_assignment"
    minimum_distance_assignment = "minimum_distance_assignment"
    sensitive_original_order_assignment = "sensitive_original_order_assignment"


class LoginResponse(BaseModel):
    access_token: Annotated[str, Field(title="Access Token")]
    refresh_token: Annotated[Optional[str], Field(title="Refresh Token")] = None
    token_type: Annotated[str, Field(title="Token Type")]


class PointOfInterest(Enum):
    """
    Available target points of interest (POI) for geolocation signal-based metrics.
    """

    start = "start"
    end = "end"


class PrivacyMetricsComputationType(Enum):
    standalone = "standalone"
    to_bottom_id_propagated = "to_bottom_id_propagated"
    to_top_enriched = "to_top_enriched"
    full_enriched = "full_enriched"


class ParameterType(Enum):
    privacy_metrics_geolocation = "privacy_metrics_geolocation"


class MetricsType(Enum):
    multi_table = "multi_table"


class MetricsType1(Enum):
    time_series = "time_series"


class PrivacyMetricsTargets(BaseModel):
    hidden_rate: Annotated[str, Field(title="Hidden Rate")]
    local_cloaking: Annotated[str, Field(title="Local Cloaking")]
    distance_to_closest: Annotated[str, Field(title="Distance To Closest")]
    closest_distances_ratio: Annotated[str, Field(title="Closest Distances Ratio")]
    column_direct_match_protection: Annotated[
        str, Field(title="Column Direct Match Protection")
    ]
    categorical_hidden_rate: Annotated[str, Field(title="Categorical Hidden Rate")]
    row_direct_match_protection: Annotated[
        str, Field(title="Row Direct Match Protection")
    ]
    correlation_protection_rate: Annotated[
        str, Field(title="Correlation Protection Rate")
    ]
    inference_continuous: Annotated[str, Field(title="Inference Continuous")]
    inference_categorical: Annotated[str, Field(title="Inference Categorical")]
    closest_rate: Annotated[str, Field(title="Closest Rate")]


class ParameterType1(Enum):
    privacy_metrics_time_series = "privacy_metrics_time_series"


class PrivacyMetricsTimeSeriesParameters(BaseModel):
    dataset_id: Annotated[UUID, Field(title="Dataset Id")]
    unshuffled_avatar_dataset_id: Annotated[
        UUID, Field(title="Unshuffled Avatar Dataset Id")
    ]
    projection_parameters: Annotated[
        Optional[Union[FPCAParameters, FlattenParameters]],
        Field(
            description="Parameters of the time series projection.",
            title="Projection Parameters",
        ),
    ] = None
    parameter_type: Annotated[ParameterType1, Field(title="Parameter Type")] = (
        ParameterType1.privacy_metrics_time_series
    )


class Projections(BaseModel):
    records: Annotated[List[List[float]], Field(title="Records")]
    avatars: Annotated[List[List[float]], Field(title="Avatars")]


class Report(BaseModel):
    id: Annotated[UUID, Field(title="Id")]
    user_id: Annotated[UUID, Field(title="User Id")]
    job_id: Annotated[UUID, Field(title="Job Id")]
    created_at: Annotated[Optional[datetime], Field(title="Created At")] = None
    download_url: Annotated[str, Field(title="Download Url")]


class ReportCreate(BaseModel):
    avatarization_job_id: Annotated[UUID, Field(title="Avatarization Job Id")]
    privacy_job_id: Annotated[UUID, Field(title="Privacy Job Id")]
    signal_job_id: Annotated[UUID, Field(title="Signal Job Id")]


class ReportFromDataCreate(BaseModel):
    dataset_id: Annotated[UUID, Field(title="Dataset Id")]
    avatars_dataset_id: Annotated[UUID, Field(title="Avatars Dataset Id")]
    privacy_job_id: Annotated[UUID, Field(title="Privacy Job Id")]
    signal_job_id: Annotated[UUID, Field(title="Signal Job Id")]


class ReportGeolocationPrivacyCreate(BaseModel):
    dataset_id: Annotated[UUID, Field(title="Dataset Id")]
    avatars_dataset_id: Annotated[UUID, Field(title="Avatars Dataset Id")]
    privacy_job_id: Annotated[UUID, Field(title="Privacy Job Id")]


class ResetPasswordRequest(BaseModel):
    email: Annotated[str, Field(title="Email")]
    new_password: Annotated[str, Field(title="New Password")]
    new_password_repeated: Annotated[str, Field(title="New Password Repeated")]
    token: Annotated[UUID, Field(title="Token")]


class SignalMetricsParameters(BaseModel):
    """
    Parameters to configure a SignalMetricsJob.
    """

    model_config = ConfigDict(
        extra="allow",
    )
    original_id: Annotated[UUID, Field(title="Original Id")]
    avatarization_job_id: Annotated[
        Optional[UUID], Field(title="Avatarization Job Id")
    ] = None
    avatars_id: Annotated[UUID, Field(title="Avatars Id")]


class SignalMetricsTargets(BaseModel):
    hellinger_mean: Annotated[str, Field(title="Hellinger Mean")]
    correlation_difference_ratio: Annotated[
        str, Field(title="Correlation Difference Ratio")
    ]


class SignalMetricsTimeSeriesParameters(BaseModel):
    dataset_id: Annotated[UUID, Field(title="Dataset Id")]
    avatar_dataset_id: Annotated[UUID, Field(title="Avatar Dataset Id")]


class SignalMetricsWithTimeSeriesParameters(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    vanilla_original_id: Annotated[
        Optional[UUID],
        Field(
            description="Identifier of the original vanilla dataset.",
            title="Vanilla Original Id",
        ),
    ] = None
    vanilla_avatars_id: Annotated[
        Optional[UUID],
        Field(
            description="Identifier of the avatar vanilla dataset.",
            title="Vanilla Avatars Id",
        ),
    ] = None
    time_series: Annotated[
        List[SignalMetricsTimeSeriesParameters],
        Field(
            description="Privacy metrics parameters of the time series. Each time series can have a different set of parameters.",
            title="Time Series",
        ),
    ]
    seed: Annotated[
        Optional[int],
        Field(description="Seed used to generate the random state.", title="Seed"),
    ] = None


class SignalPosition(Enum):
    """
    Available known signal positions for geolocation signal-based metrics.
    """

    start = "start"
    middle = "middle"
    end = "end"


class TableReference(BaseModel):
    dataset_id: Annotated[UUID, Field(title="Dataset Id")]
    is_individual_level: Annotated[
        bool,
        Field(
            description="Whether the table contains rows at the individual level. Individual level means that each row contains information about a distinct physical person.",
            title="Is Individual Level",
        ),
    ]


class TimeSeriesSignalMetricsPerDataset(BaseModel):
    pointwise_hellinger_mean: Annotated[float, Field(title="Pointwise Hellinger Mean")]
    pointwise_hellinger_std: Annotated[float, Field(title="Pointwise Hellinger Std")]
    density_hellinger_mean: Annotated[float, Field(title="Density Hellinger Mean")]
    density_hellinger_std: Annotated[float, Field(title="Density Hellinger Std")]
    original_id: Annotated[UUID, Field(title="Original Id")]
    avatars_id: Annotated[UUID, Field(title="Avatars Id")]


class UserRole(Enum):
    admin = "admin"
    user = "user"


class ValidationError(BaseModel):
    loc: Annotated[List[Union[str, int]], Field(title="Location")]
    msg: Annotated[str, Field(title="Message")]
    type: Annotated[str, Field(title="Error Type")]


class GrantTypeItem(RootModel[str]):
    root: Annotated[str, Field(pattern="password")]


class Login(BaseModel):
    grant_type: Annotated[Optional[GrantTypeItem], Field(title="Grant Type")] = None
    username: Annotated[str, Field(title="Username")]
    password: Annotated[str, Field(title="Password")]
    scope: Annotated[Optional[str], Field(title="Scope")] = ""
    client_id: Annotated[Optional[str], Field(title="Client Id")] = None
    client_secret: Annotated[Optional[str], Field(title="Client Secret")] = None


class CreateDataset(BaseModel):
    file: Annotated[bytes, Field(title="File")]
    name: Annotated[Optional[str], Field(title="Name")] = None


class GenericJob(BaseModel):
    id: Annotated[UUID, Field(title="Id")]
    kind: JobKind
    created_at: Annotated[datetime, Field(title="Created At")]
    status: JobStatus
    error_message: Annotated[Optional[str], Field(title="Error Message")] = None
    traceback: Annotated[Optional[str], Field(title="Traceback")] = None
    result: Optional[GenericResult] = None
    parameters: GenericParameters
    current_progress: Optional[JobProgress] = None


class AdviceJobCreate(BaseModel):
    parameters: AdviceParameters


class AvatarizationTimeSeriesParameters(BaseModel):
    """
    Parameters for the anonymization of a Time Series.
    """

    dataset_id: Annotated[
        UUID,
        Field(
            description="Identifier of the dataset containing the time series.",
            title="Dataset Id",
        ),
    ]
    projection_parameters: Annotated[
        Union[FPCAParameters, FlattenParameters],
        Field(
            description="Parameters of the time series projection.",
            title="Projection Parameters",
        ),
    ]


class ColumnDetail(BaseModel):
    type: ColumnType
    label: Annotated[str, Field(title="Label")]
    is_identifier: Annotated[Optional[bool], Field(title="Is Identifier")] = False


class CompatibilityResponse(BaseModel):
    message: Annotated[str, Field(title="Message")]
    most_recent_compatible_client: Annotated[
        Optional[str], Field(title="Most Recent Compatible Client")
    ] = None
    status: CompatibilityStatus


class CreateUser(BaseModel):
    """
    Create a user, either with an email, or a username.

    The choice will depend on how your server is setup.
    """

    username: Annotated[Optional[str], Field(title="Username")] = None
    email: Annotated[Optional[str], Field(title="Email")] = None
    role: Optional[UserRole] = UserRole.user
    password: Annotated[Optional[str], Field(title="Password")] = None


class Dataset(BaseModel):
    id: Annotated[UUID, Field(title="Id")]
    hash: Annotated[str, Field(title="Hash")]
    name: Annotated[Optional[str], Field(title="Name")] = None
    columns: Annotated[List[ColumnDetail], Field(title="Columns")]
    download_url: Annotated[str, Field(title="Download Url")]
    analysis_status: Optional[AnalysisStatus] = None
    analysis_duration: Annotated[Optional[float], Field(title="Analysis Duration")] = (
        None
    )
    nb_lines: Annotated[Optional[int], Field(title="Nb Lines")] = None
    nb_dimensions: Annotated[int, Field(title="Nb Dimensions")]
    summary: Optional[DatasetSummary] = None
    filetype: FileType


class HTTPValidationError(BaseModel):
    detail: Annotated[Optional[List[ValidationError]], Field(title="Detail")] = None


class ImputationParameters(BaseModel):
    method: Optional[ImputeMethod] = None
    k: Annotated[Optional[int], Field(title="K")] = None
    training_fraction: Annotated[
        Optional[TrainingFractionItem], Field(title="Training Fraction")
    ] = None


class PrivacyMetrics(BaseModel):
    hidden_rate: Annotated[float, Field(title="Hidden Rate")]
    local_cloaking: Annotated[float, Field(title="Local Cloaking")]
    distance_to_closest: Annotated[
        Optional[float], Field(title="Distance To Closest")
    ] = None
    closest_distances_ratio: Annotated[
        Optional[float], Field(title="Closest Distances Ratio")
    ] = None
    column_direct_match_protection: Annotated[
        float, Field(title="Column Direct Match Protection")
    ]
    categorical_hidden_rate: Annotated[
        Optional[float], Field(title="Categorical Hidden Rate")
    ] = None
    row_direct_match_protection: Annotated[
        float, Field(title="Row Direct Match Protection")
    ]
    correlation_protection_rate: Annotated[
        Optional[float], Field(title="Correlation Protection Rate")
    ] = None
    inference_continuous: Annotated[
        Optional[float], Field(title="Inference Continuous")
    ] = None
    inference_categorical: Annotated[
        Optional[float], Field(title="Inference Categorical")
    ] = None
    closest_rate: Annotated[Optional[float], Field(title="Closest Rate")] = None
    targets: PrivacyMetricsTargets


class PrivacyMetricsGeolocationScenario(BaseModel):
    projection_parameters: Annotated[
        Union[GeolocationDensityParameters, GeolocationFeaturesParameters],
        Field(
            description="Parameters of the geolocation projection.",
            title="Projection Parameters",
        ),
    ]
    known_features: Annotated[
        Optional[List[GeolocationFeatures]],
        Field(description="Known geolocation features", title="Known Features"),
    ] = None
    target_feature: Annotated[
        Optional[GeolocationFeatures], Field(description="Target geolocation features")
    ] = None
    known_signal_position: Annotated[
        Optional[SignalPosition], Field(description="Position of the known signal")
    ] = None
    target_signal_poi: Annotated[
        Optional[PointOfInterest],
        Field(description="Target point of interest (POI) to infer"),
    ] = None
    inference_metric_threshold: Annotated[
        Optional[float],
        Field(
            description="Threshold (in meters) for the inference metric",
            title="Inference Metric Threshold",
        ),
    ] = 500


class PrivacyMetricsGeolocationScenarioResult(BaseModel):
    privacy_metrics: GeolocationPrivacyMetrics
    parameters: PrivacyMetricsGeolocationScenario


class PrivacyMetricsParameters(BaseModel):
    """
    Parameters to configure a PrivacyMetricsJob.
    """

    model_config = ConfigDict(
        extra="allow",
    )
    original_id: Annotated[UUID, Field(title="Original Id")]
    avatarization_job_id: Annotated[
        Optional[UUID], Field(title="Avatarization Job Id")
    ] = None
    imputation: Annotated[
        Optional[ImputationParameters],
        Field(
            description="Imputation parameters used to preprocess avatars data for metrics computation."
        ),
    ] = None
    ncp: Annotated[
        Optional[int],
        Field(
            description="Number of components used for distance based metrics computation. Default behavior similar to avatarization. If unspecified, ncp is set to the minimum value between 5 and the number of variables after one hot encoding. Note: we recommend using to use the same ncp as during the avatarization to check the worst-case scenario.",
            title="Ncp",
        ),
    ] = None
    use_categorical_reduction: Annotated[
        Optional[bool],
        Field(
            description="Parameter to vectorize categorical variables using cat2vec. See :class:`AvatarizationParameters` for more details. Note: we recommend using the same parameter as during the avatarization.",
            title="Use Categorical Reduction",
        ),
    ] = None
    known_variables: Annotated[
        Optional[List[str]],
        Field(
            description="Variables considered to be known by the attacker. Used to measure some specific privacy metrics. Note: we recommend using demographic variables as they are the most likely to be known by an attacker.",
            title="Known Variables",
        ),
    ] = None
    target: Annotated[
        Optional[str],
        Field(
            description="The target variable to be inferred. Used to measure some specific privacy metrics. Must be used together with ``known_variables``. Note: we recommend using a sensitive variable",
            title="Target",
        ),
    ] = None
    closest_rate_percentage_threshold: Annotated[
        Optional[ClosestRatePercentageThresholdItem],
        Field(
            description="Threshold of closest original individuals to define a distance threshold. Any avatar data generated closer from an original than the distance threshold, is considered at risk.",
            title="Closest Rate Percentage Threshold",
        ),
    ] = None
    closest_rate_ratio_threshold: Annotated[
        Optional[ClosestRateRatioThresholdItem],
        Field(
            description="Closest distance ratio threshold below which an avatar data is considered too close to an original and too isolated from the rest to be safe.",
            title="Closest Rate Ratio Threshold",
        ),
    ] = None
    categorical_hidden_rate_variables: Annotated[
        Optional[List[str]],
        Field(
            description="List of categorical variables to compute the categorical hidden rate.",
            title="Categorical Hidden Rate Variables",
        ),
    ] = []
    unshuffled_avatars_id: Annotated[UUID, Field(title="Unshuffled Avatars Id")]
    seed: Annotated[
        Optional[int],
        Field(description="Seed used to generate the random state.", title="Seed"),
    ] = None


class PrivacyMetricsPerMultiTableDataset(BaseModel):
    hidden_rate: Annotated[float, Field(title="Hidden Rate")]
    local_cloaking: Annotated[float, Field(title="Local Cloaking")]
    distance_to_closest: Annotated[
        Optional[float], Field(title="Distance To Closest")
    ] = None
    closest_distances_ratio: Annotated[
        Optional[float], Field(title="Closest Distances Ratio")
    ] = None
    column_direct_match_protection: Annotated[
        float, Field(title="Column Direct Match Protection")
    ]
    categorical_hidden_rate: Annotated[
        Optional[float], Field(title="Categorical Hidden Rate")
    ] = None
    row_direct_match_protection: Annotated[
        float, Field(title="Row Direct Match Protection")
    ]
    correlation_protection_rate: Annotated[
        Optional[float], Field(title="Correlation Protection Rate")
    ] = None
    inference_continuous: Annotated[
        Optional[float], Field(title="Inference Continuous")
    ] = None
    inference_categorical: Annotated[
        Optional[float], Field(title="Inference Categorical")
    ] = None
    closest_rate: Annotated[Optional[float], Field(title="Closest Rate")] = None
    targets: PrivacyMetricsTargets
    original_id: Annotated[UUID, Field(title="Original Id")]
    unshuffled_avatars_id: Annotated[UUID, Field(title="Unshuffled Avatars Id")]
    metrics_type: Annotated[MetricsType, Field(title="Metrics Type")] = (
        MetricsType.multi_table
    )
    individual_identifier_variable: Annotated[
        str, Field(title="Individual Identifier Variable")
    ]
    computation_type: PrivacyMetricsComputationType
    dataset_name: Annotated[str, Field(title="Dataset Name")]


class PrivacyMetricsPerTimeSeriesDataset(BaseModel):
    hidden_rate: Annotated[float, Field(title="Hidden Rate")]
    local_cloaking: Annotated[float, Field(title="Local Cloaking")]
    distance_to_closest: Annotated[
        Optional[float], Field(title="Distance To Closest")
    ] = None
    closest_distances_ratio: Annotated[
        Optional[float], Field(title="Closest Distances Ratio")
    ] = None
    column_direct_match_protection: Annotated[
        float, Field(title="Column Direct Match Protection")
    ]
    categorical_hidden_rate: Annotated[
        Optional[float], Field(title="Categorical Hidden Rate")
    ] = None
    row_direct_match_protection: Annotated[
        float, Field(title="Row Direct Match Protection")
    ]
    correlation_protection_rate: Annotated[
        Optional[float], Field(title="Correlation Protection Rate")
    ] = None
    inference_continuous: Annotated[
        Optional[float], Field(title="Inference Continuous")
    ] = None
    inference_categorical: Annotated[
        Optional[float], Field(title="Inference Categorical")
    ] = None
    closest_rate: Annotated[Optional[float], Field(title="Closest Rate")] = None
    targets: PrivacyMetricsTargets
    original_id: Annotated[UUID, Field(title="Original Id")]
    unshuffled_avatars_id: Annotated[UUID, Field(title="Unshuffled Avatars Id")]
    metrics_type: Annotated[MetricsType1, Field(title="Metrics Type")] = (
        MetricsType1.time_series
    )


class PrivacyMetricsWithTimeSeriesParameters(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    imputation: Annotated[
        Optional[ImputationParameters],
        Field(
            description="Imputation parameters used to preprocess avatars data for metrics computation."
        ),
    ] = None
    ncp: Annotated[
        Optional[int],
        Field(
            description="Number of components used for distance based metrics computation. Default behavior similar to avatarization. If unspecified, ncp is set to the minimum value between 5 and the number of variables after one hot encoding. Note: we recommend using to use the same ncp as during the avatarization to check the worst-case scenario.",
            title="Ncp",
        ),
    ] = None
    use_categorical_reduction: Annotated[
        Optional[bool],
        Field(
            description="Parameter to vectorize categorical variables using cat2vec. See :class:`AvatarizationParameters` for more details. Note: we recommend using the same parameter as during the avatarization.",
            title="Use Categorical Reduction",
        ),
    ] = None
    known_variables: Annotated[
        Optional[List[str]],
        Field(
            description="Variables considered to be known by the attacker. Used to measure some specific privacy metrics. Note: we recommend using demographic variables as they are the most likely to be known by an attacker.",
            title="Known Variables",
        ),
    ] = None
    target: Annotated[
        Optional[str],
        Field(
            description="The target variable to be inferred. Used to measure some specific privacy metrics. Must be used together with ``known_variables``. Note: we recommend using a sensitive variable",
            title="Target",
        ),
    ] = None
    closest_rate_percentage_threshold: Annotated[
        Optional[ClosestRatePercentageThresholdItem],
        Field(
            description="Threshold of closest original individuals to define a distance threshold. Any avatar data generated closer from an original than the distance threshold, is considered at risk.",
            title="Closest Rate Percentage Threshold",
        ),
    ] = None
    closest_rate_ratio_threshold: Annotated[
        Optional[ClosestRateRatioThresholdItem],
        Field(
            description="Closest distance ratio threshold below which an avatar data is considered too close to an original and too isolated from the rest to be safe.",
            title="Closest Rate Ratio Threshold",
        ),
    ] = None
    categorical_hidden_rate_variables: Annotated[
        Optional[List[str]],
        Field(
            description="List of categorical variables to compute the categorical hidden rate.",
            title="Categorical Hidden Rate Variables",
        ),
    ] = []
    vanilla_original_id: Annotated[
        Optional[UUID],
        Field(
            description="Identifier of the original vanilla dataset.",
            title="Vanilla Original Id",
        ),
    ] = None
    vanilla_unshuffled_avatars_id: Annotated[
        Optional[UUID],
        Field(
            description="Identifier of the avatar vanilla dataset.",
            title="Vanilla Unshuffled Avatars Id",
        ),
    ] = None
    time_series: Annotated[
        List[PrivacyMetricsTimeSeriesParameters],
        Field(
            description="Privacy metrics parameters of the time series. Each time series can have a different set of parameters.",
            title="Time Series",
        ),
    ]
    seed: Annotated[
        Optional[int],
        Field(description="Seed used to generate the random state.", title="Seed"),
    ] = None


class SignalMetrics(BaseModel):
    hellinger_mean: Annotated[float, Field(title="Hellinger Mean")]
    hellinger_std: Annotated[float, Field(title="Hellinger Std")]
    correlation_difference_ratio: Annotated[
        Optional[float], Field(title="Correlation Difference Ratio")
    ] = None
    targets: SignalMetricsTargets


class SignalMetricsJobCreate(BaseModel):
    parameters: SignalMetricsParameters


class SignalMetricsPerDataset(BaseModel):
    hellinger_mean: Annotated[float, Field(title="Hellinger Mean")]
    hellinger_std: Annotated[float, Field(title="Hellinger Std")]
    correlation_difference_ratio: Annotated[
        Optional[float], Field(title="Correlation Difference Ratio")
    ] = None
    targets: SignalMetricsTargets
    original_id: Annotated[UUID, Field(title="Original Id")]
    avatars_id: Annotated[UUID, Field(title="Avatars Id")]


class SignalMetricsWithTimeSeriesJobCreate(BaseModel):
    parameters: SignalMetricsWithTimeSeriesParameters


class TableLink(BaseModel):
    parent_table: Annotated[
        TableReference, Field(description="Reference to the parent table.")
    ]
    child_table: Annotated[
        TableReference, Field(description="Reference to the child table.")
    ]
    parent_link_key: Annotated[
        str,
        Field(
            description="Variable to be used for linking with the child table. The link_key specified here must contain unique values only. It is most often the\n        primary key. It needs to have the is_identifier property set to True when creating\n        the dataset.",
            title="Parent Link Key",
        ),
    ]
    child_link_key: Annotated[
        str,
        Field(
            description="Variable to be used for linking with the parent table. It is most often a foreign key. It needs to have the property is_identifier set to True\n        when creating the dataset.",
            title="Child Link Key",
        ),
    ]
    link_method: Annotated[
        Optional[LinkMethod], Field(description="Assignment method.")
    ] = LinkMethod.linear_sum_assignment


class User(BaseModel):
    id: Annotated[UUID, Field(title="Id")]
    organization_id: Annotated[UUID, Field(title="Organization Id")]
    username: Annotated[Optional[str], Field(title="Username")] = None
    email: Annotated[Optional[str], Field(title="Email")] = None
    role: Optional[UserRole] = UserRole.user


class PatchDataset(BaseModel):
    columns: Annotated[Optional[List[ColumnDetail]], Field(title="Columns")] = None


class PrivacyMetricsJob(BaseModel):
    id: Annotated[UUID, Field(title="Id")]
    kind: JobKind
    created_at: Annotated[datetime, Field(title="Created At")]
    status: JobStatus
    error_message: Annotated[Optional[str], Field(title="Error Message")] = None
    traceback: Annotated[Optional[str], Field(title="Traceback")] = None
    result: Optional[PrivacyMetrics] = None
    parameters: PrivacyMetricsParameters
    current_progress: Optional[JobProgress] = None


class SignalMetricsJob(BaseModel):
    id: Annotated[UUID, Field(title="Id")]
    kind: JobKind
    created_at: Annotated[datetime, Field(title="Created At")]
    status: JobStatus
    error_message: Annotated[Optional[str], Field(title="Error Message")] = None
    traceback: Annotated[Optional[str], Field(title="Traceback")] = None
    result: Optional[SignalMetrics] = None
    parameters: SignalMetricsParameters
    current_progress: Optional[JobProgress] = None


class AvatarizationParameters(BaseModel):
    """
    Parameters to create an anonymization job.

    It contains the seed parameter.
    """

    model_config = ConfigDict(
        extra="allow",
    )
    k: Annotated[int, Field(title="K")]
    column_weights: Annotated[
        Optional[Dict[str, float]], Field(title="Column Weights")
    ] = None
    ncp: Annotated[Optional[int], Field(title="Ncp")] = None
    imputation: Optional[ImputationParameters] = None
    use_categorical_reduction: Annotated[
        Optional[bool], Field(title="Use Categorical Reduction")
    ] = None
    exclude_variables: Optional[ExcludeVariablesParameters] = None
    dataset_id: Annotated[UUID, Field(title="Dataset Id")]
    seed: Annotated[Optional[int], Field(title="Seed")] = None


class AvatarizationResult(BaseModel):
    privacy_metrics: Optional[PrivacyMetrics] = None
    signal_metrics: Optional[SignalMetrics] = None
    avatars_dataset: Dataset
    sensitive_unshuffled_avatars_datasets: Dataset


class AvatarizationResultPerDataset(BaseModel):
    privacy_metrics: Optional[PrivacyMetrics] = None
    signal_metrics: Optional[SignalMetrics] = None
    avatars_dataset: Dataset
    sensitive_unshuffled_avatars_datasets: Dataset
    original_id: Annotated[UUID, Field(title="Original Id")]


class AvatarizationWithTimeSeriesParameters(BaseModel):
    """
    Parameters to create an time series anonymization job.

    It contains the seed parameter.
    """

    model_config = ConfigDict(
        extra="allow",
    )
    k: Annotated[int, Field(title="K")]
    column_weights: Annotated[
        Optional[Dict[str, float]], Field(title="Column Weights")
    ] = None
    ncp: Annotated[Optional[int], Field(title="Ncp")] = None
    imputation: Optional[ImputationParameters] = None
    use_categorical_reduction: Annotated[
        Optional[bool], Field(title="Use Categorical Reduction")
    ] = None
    exclude_variables: Optional[ExcludeVariablesParameters] = None
    vanilla_dataset_id: Annotated[
        Optional[UUID],
        Field(
            description="Dataset identifier of the vanilla data. ",
            title="Vanilla Dataset Id",
        ),
    ] = None
    time_series: Annotated[
        List[AvatarizationTimeSeriesParameters],
        Field(
            description="Parameters of the time series. Each time series can have a different set of parameters.",
            title="Time Series",
        ),
    ]
    seed: Annotated[Optional[int], Field(title="Seed")] = None


class AvatarizationWithTimeSeriesResult(BaseModel):
    datasets: Annotated[List[AvatarizationResultPerDataset], Field(title="Datasets")]


class BaseAvatarizationParameters(BaseModel):
    """
    Parameters to create an anonymization with vanilla dataset.

    This class is used in the multi table job. It does not contain the seed.
    """

    k: Annotated[int, Field(title="K")]
    column_weights: Annotated[
        Optional[Dict[str, float]], Field(title="Column Weights")
    ] = None
    ncp: Annotated[Optional[int], Field(title="Ncp")] = None
    imputation: Optional[ImputationParameters] = None
    use_categorical_reduction: Annotated[
        Optional[bool], Field(title="Use Categorical Reduction")
    ] = None
    exclude_variables: Optional[ExcludeVariablesParameters] = None
    dataset_id: Annotated[UUID, Field(title="Dataset Id")]


class BaseAvatarizationWithTimeSeriesParameters(BaseModel):
    """
    Parameters to configure an anonymization with a time series dataset.

    This class is used in the multi table job. It does not contain the seed.
    """

    k: Annotated[int, Field(title="K")]
    column_weights: Annotated[
        Optional[Dict[str, float]], Field(title="Column Weights")
    ] = None
    ncp: Annotated[Optional[int], Field(title="Ncp")] = None
    imputation: Optional[ImputationParameters] = None
    use_categorical_reduction: Annotated[
        Optional[bool], Field(title="Use Categorical Reduction")
    ] = None
    exclude_variables: Optional[ExcludeVariablesParameters] = None
    vanilla_dataset_id: Annotated[
        Optional[UUID],
        Field(
            description="Dataset identifier of the vanilla data. ",
            title="Vanilla Dataset Id",
        ),
    ] = None
    time_series: Annotated[
        List[AvatarizationTimeSeriesParameters],
        Field(
            description="Parameters of the time series. Each time series can have a different set of parameters.",
            title="Time Series",
        ),
    ]


class BasePrivacyMetricsParameters(BaseModel):
    """
    Parameters to configure privacy metrics.

    This class is used in the multi table job. It does not contain the seed.
    """

    model_config = ConfigDict(
        extra="allow",
    )
    original_id: Annotated[UUID, Field(title="Original Id")]
    avatarization_job_id: Annotated[
        Optional[UUID], Field(title="Avatarization Job Id")
    ] = None
    imputation: Annotated[
        Optional[ImputationParameters],
        Field(
            description="Imputation parameters used to preprocess avatars data for metrics computation."
        ),
    ] = None
    ncp: Annotated[
        Optional[int],
        Field(
            description="Number of components used for distance based metrics computation. Default behavior similar to avatarization. If unspecified, ncp is set to the minimum value between 5 and the number of variables after one hot encoding. Note: we recommend using to use the same ncp as during the avatarization to check the worst-case scenario.",
            title="Ncp",
        ),
    ] = None
    use_categorical_reduction: Annotated[
        Optional[bool],
        Field(
            description="Parameter to vectorize categorical variables using cat2vec. See :class:`AvatarizationParameters` for more details. Note: we recommend using the same parameter as during the avatarization.",
            title="Use Categorical Reduction",
        ),
    ] = None
    known_variables: Annotated[
        Optional[List[str]],
        Field(
            description="Variables considered to be known by the attacker. Used to measure some specific privacy metrics. Note: we recommend using demographic variables as they are the most likely to be known by an attacker.",
            title="Known Variables",
        ),
    ] = None
    target: Annotated[
        Optional[str],
        Field(
            description="The target variable to be inferred. Used to measure some specific privacy metrics. Must be used together with ``known_variables``. Note: we recommend using a sensitive variable",
            title="Target",
        ),
    ] = None
    closest_rate_percentage_threshold: Annotated[
        Optional[ClosestRatePercentageThresholdItem],
        Field(
            description="Threshold of closest original individuals to define a distance threshold. Any avatar data generated closer from an original than the distance threshold, is considered at risk.",
            title="Closest Rate Percentage Threshold",
        ),
    ] = None
    closest_rate_ratio_threshold: Annotated[
        Optional[ClosestRateRatioThresholdItem],
        Field(
            description="Closest distance ratio threshold below which an avatar data is considered too close to an original and too isolated from the rest to be safe.",
            title="Closest Rate Ratio Threshold",
        ),
    ] = None
    categorical_hidden_rate_variables: Annotated[
        Optional[List[str]],
        Field(
            description="List of categorical variables to compute the categorical hidden rate.",
            title="Categorical Hidden Rate Variables",
        ),
    ] = []
    unshuffled_avatars_id: Annotated[UUID, Field(title="Unshuffled Avatars Id")]


class BasePrivacyMetricsWithTimeSeriesParameters(BaseModel):
    """
    Parameters to configure a privacy metric job with a time series dataset.

    This class is used in the multi table job. It does not contain the seed.
    """

    model_config = ConfigDict(
        extra="allow",
    )
    imputation: Annotated[
        Optional[ImputationParameters],
        Field(
            description="Imputation parameters used to preprocess avatars data for metrics computation."
        ),
    ] = None
    ncp: Annotated[
        Optional[int],
        Field(
            description="Number of components used for distance based metrics computation. Default behavior similar to avatarization. If unspecified, ncp is set to the minimum value between 5 and the number of variables after one hot encoding. Note: we recommend using to use the same ncp as during the avatarization to check the worst-case scenario.",
            title="Ncp",
        ),
    ] = None
    use_categorical_reduction: Annotated[
        Optional[bool],
        Field(
            description="Parameter to vectorize categorical variables using cat2vec. See :class:`AvatarizationParameters` for more details. Note: we recommend using the same parameter as during the avatarization.",
            title="Use Categorical Reduction",
        ),
    ] = None
    known_variables: Annotated[
        Optional[List[str]],
        Field(
            description="Variables considered to be known by the attacker. Used to measure some specific privacy metrics. Note: we recommend using demographic variables as they are the most likely to be known by an attacker.",
            title="Known Variables",
        ),
    ] = None
    target: Annotated[
        Optional[str],
        Field(
            description="The target variable to be inferred. Used to measure some specific privacy metrics. Must be used together with ``known_variables``. Note: we recommend using a sensitive variable",
            title="Target",
        ),
    ] = None
    closest_rate_percentage_threshold: Annotated[
        Optional[ClosestRatePercentageThresholdItem],
        Field(
            description="Threshold of closest original individuals to define a distance threshold. Any avatar data generated closer from an original than the distance threshold, is considered at risk.",
            title="Closest Rate Percentage Threshold",
        ),
    ] = None
    closest_rate_ratio_threshold: Annotated[
        Optional[ClosestRateRatioThresholdItem],
        Field(
            description="Closest distance ratio threshold below which an avatar data is considered too close to an original and too isolated from the rest to be safe.",
            title="Closest Rate Ratio Threshold",
        ),
    ] = None
    categorical_hidden_rate_variables: Annotated[
        Optional[List[str]],
        Field(
            description="List of categorical variables to compute the categorical hidden rate.",
            title="Categorical Hidden Rate Variables",
        ),
    ] = []
    vanilla_original_id: Annotated[
        Optional[UUID],
        Field(
            description="Identifier of the original vanilla dataset.",
            title="Vanilla Original Id",
        ),
    ] = None
    vanilla_unshuffled_avatars_id: Annotated[
        Optional[UUID],
        Field(
            description="Identifier of the avatar vanilla dataset.",
            title="Vanilla Unshuffled Avatars Id",
        ),
    ] = None
    time_series: Annotated[
        List[PrivacyMetricsTimeSeriesParameters],
        Field(
            description="Privacy metrics parameters of the time series. Each time series can have a different set of parameters.",
            title="Time Series",
        ),
    ]


class MetaPrivacyMetrics(BaseModel):
    details: Annotated[
        List[
            Union[
                PrivacyMetricsPerMultiTableDataset, PrivacyMetricsPerTimeSeriesDataset
            ]
        ],
        Field(title="Details"),
    ]


class MetaSignalMetrics(BaseModel):
    details: Annotated[
        List[Union[SignalMetricsPerDataset, TimeSeriesSignalMetricsPerDataset]],
        Field(title="Details"),
    ]


class PrivacyMetricsGeolocationParameters(BaseModel):
    """
    Parameters to configure a PrivacyMetricsGeolocationJob.

    It enables to compute privacy metrics for multiple scenarios (with no limit in number).
    """

    model_config = ConfigDict(
        extra="allow",
    )
    original_dataset_id: Annotated[UUID, Field(title="Original Dataset Id")]
    unshuffled_avatar_dataset_id: Annotated[
        UUID, Field(title="Unshuffled Avatar Dataset Id")
    ]
    scenarios: Annotated[
        List[PrivacyMetricsGeolocationScenario], Field(title="Scenarios")
    ]
    parameter_type: Annotated[ParameterType, Field(title="Parameter Type")] = (
        ParameterType.privacy_metrics_geolocation
    )


class PrivacyMetricsGeolocationResult(BaseModel):
    privacy_metrics_per_scenario: Annotated[
        List[PrivacyMetricsGeolocationScenarioResult],
        Field(title="Privacy Metrics Per Scenario"),
    ]


class PrivacyMetricsJobCreate(BaseModel):
    parameters: PrivacyMetricsParameters


class PrivacyMetricsMultiTableParameters(BaseModel):
    """
    Parameters to create a multi table privacy metrics job.

    There are a few rules to follow to make sure that the parameters are consistent.
    The following ones are the same as the ones for the avatarization job.
    Other rules specific to privacy metrics are down below.

    - Every dataset_id in table_parameters must be referenced in table_links.
    - Every dataset_id in table_links must also be referenced in table_parameters.
    - Every TableReference with the same dataset_id must be identical in other fields.
    - There can't be a cycle in the table_links e.g. A -> B -> C -> A is not allowed.
    - Each table needs to have N+1 identifier columns (id, patient_id, visit_id, etc.), where N is
      the  number of direct predecessors of each table. A direct predecessor is a table that has a
      direct link (depth=1) between itself and another table.

      For example, if a table has 2 direct predecessors, it needs to have 3 identifier columns.

      Examples:
        * A -> B <- C , A and C need to have 1 identifier column, and B needs to have 3 identifier
            columns.
        * A -> B -> C <- D , A and D need to have 1 identifier column, B needs to have 2, and
         C needs to only have 3 and not 4, because C only has 2 direct predecessors (B and D).
         There is not a direct link from C to A, so it is not a direct predecessor.
    - Each table can only be linked to the same variable once. This means that if A -> B <- C ,
      A and C need to link to two different columns in B.
    - Every table can only reference another table once. This means that A -> B <- A is
      not possible.
    - Top level tables need to be individual_level tables. For example, A -> B <- C , A and C need
      to be individual_level tables.

      Privacy Metrics specific
      ------------------------

      - The avatar datasets referenced by the avatars_id in table_parameters need to be
        identical in columns, shape, and dtype to their corresponding original dataset.
    """

    model_config = ConfigDict(
        extra="allow",
    )
    table_links: Annotated[
        List[TableLink],
        Field(
            description="Description of each relational link between tables.",
            min_length=1,
            title="Table Links",
        ),
    ]
    table_parameters: Annotated[
        List[
            Union[
                BasePrivacyMetricsParameters, BasePrivacyMetricsWithTimeSeriesParameters
            ]
        ],
        Field(
            description="Privacy metrics parameters for each table. ",
            min_length=1,
            title="Table Parameters",
        ),
    ]
    seed: Annotated[
        Optional[int],
        Field(description="Seed used to generate the random state.", title="Seed"),
    ] = None


class PrivacyMetricsWithTimeSeriesJobCreate(BaseModel):
    parameters: PrivacyMetricsWithTimeSeriesParameters


class AvatarizationJob(BaseModel):
    id: Annotated[UUID, Field(title="Id")]
    kind: JobKind
    created_at: Annotated[datetime, Field(title="Created At")]
    status: JobStatus
    error_message: Annotated[Optional[str], Field(title="Error Message")] = None
    traceback: Annotated[Optional[str], Field(title="Traceback")] = None
    result: Optional[AvatarizationResult] = None
    parameters: AvatarizationParameters
    current_progress: Optional[JobProgress] = None


class AvatarizationWithTimeSeriesJob(BaseModel):
    id: Annotated[UUID, Field(title="Id")]
    kind: JobKind
    created_at: Annotated[datetime, Field(title="Created At")]
    status: JobStatus
    error_message: Annotated[Optional[str], Field(title="Error Message")] = None
    traceback: Annotated[Optional[str], Field(title="Traceback")] = None
    result: Optional[AvatarizationWithTimeSeriesResult] = None
    parameters: AvatarizationWithTimeSeriesParameters
    current_progress: Optional[JobProgress] = None


class PrivacyMetricsGeolocationJob(BaseModel):
    id: Annotated[UUID, Field(title="Id")]
    kind: JobKind
    created_at: Annotated[datetime, Field(title="Created At")]
    status: JobStatus
    error_message: Annotated[Optional[str], Field(title="Error Message")] = None
    traceback: Annotated[Optional[str], Field(title="Traceback")] = None
    result: Optional[PrivacyMetricsGeolocationResult] = None
    parameters: PrivacyMetricsGeolocationParameters
    current_progress: Optional[JobProgress] = None


class PrivacyMetricsMultiTableJob(BaseModel):
    id: Annotated[UUID, Field(title="Id")]
    kind: JobKind
    created_at: Annotated[datetime, Field(title="Created At")]
    status: JobStatus
    error_message: Annotated[Optional[str], Field(title="Error Message")] = None
    traceback: Annotated[Optional[str], Field(title="Traceback")] = None
    result: Optional[MetaPrivacyMetrics] = None
    parameters: PrivacyMetricsMultiTableParameters
    current_progress: Optional[JobProgress] = None


class PrivacyMetricsWithTimeSeriesJob(BaseModel):
    id: Annotated[UUID, Field(title="Id")]
    kind: JobKind
    created_at: Annotated[datetime, Field(title="Created At")]
    status: JobStatus
    error_message: Annotated[Optional[str], Field(title="Error Message")] = None
    traceback: Annotated[Optional[str], Field(title="Traceback")] = None
    result: Optional[MetaPrivacyMetrics] = None
    parameters: PrivacyMetricsWithTimeSeriesParameters
    current_progress: Optional[JobProgress] = None


class SignalMetricsWithTimeSeriesJob(BaseModel):
    id: Annotated[UUID, Field(title="Id")]
    kind: JobKind
    created_at: Annotated[datetime, Field(title="Created At")]
    status: JobStatus
    error_message: Annotated[Optional[str], Field(title="Error Message")] = None
    traceback: Annotated[Optional[str], Field(title="Traceback")] = None
    result: Optional[MetaSignalMetrics] = None
    parameters: SignalMetricsWithTimeSeriesParameters
    current_progress: Optional[JobProgress] = None


class Advice(BaseModel):
    """
    Recommendations on anonymization parameters and processors.
    """

    parameters: Annotated[
        AvatarizationParameters, Field(description="Suggested AvatarizationParameters.")
    ]
    python_client_pipeline: Annotated[
        str,
        Field(
            description="Pipeline ready to run, you need to replace <NAME_OF_YOUR_DF> by the name of your dataframe.",
            title="Python Client Pipeline",
        ),
    ]
    more_details: Annotated[
        Optional[Dict[str, str]],
        Field(
            description="Additional information on parameters and processors.",
            title="More Details",
        ),
    ] = {}


class AvatarizationJobCreate(BaseModel):
    kind: Optional[JobKind] = JobKind.avatarization
    parameters: AvatarizationParameters


class AvatarizationMultiTableParameters(BaseModel):
    """
    Parameters to create a multi table anonymization job.

    There are a few rules to follow to make sure that the parameters are consistent:

    - Every dataset_id in table_parameters must be referenced in table_links.
    - Every dataset_id in table_links must also be referenced in table_parameters.
    - Every TableReference with the same dataset_id must be identical in other fields.
    - There can't be a cycle in the table_links e.g. A -> B -> C -> A is not allowed.
    - Each table needs to have N+1 identifier columns (id, patient_id, visit_id, etc.), where N is
      the  number of direct predecessors of each table. A direct predecessor is a table that has a
      direct link (depth=1) between itself and another table.

      For example, if a table has 2 direct predecessors, it needs to have 3 identifier columns.

      Examples:
        * A -> B <- C , A and C need to have 1 identifier column, and B needs to have 3 identifier
            columns.
        * A -> B -> C <- D , A and D need to have 1 identifier column, B needs to have 2, and
         C needs to only have 3 and not 4, because C only has 2 direct predecessors (B and D).
         There is not a direct link from C to A, so it is not a direct predecessor.
    - Each table can only be linked to the same variable once. This means that if A -> B <- C ,
      A and C need to link to two different columns in B.
    - Every table can only reference another table once. This means that A -> B <- A is
      not possible.
    - Top level tables need to be individual_level tables. For example, A -> B <- C , A and C need
      to be individual_level tables.
    """

    model_config = ConfigDict(
        extra="allow",
    )
    table_links: Annotated[
        List[TableLink],
        Field(
            description="Description of each relational link between tables.",
            min_length=1,
            title="Table Links",
        ),
    ]
    table_parameters: Annotated[
        List[
            Union[
                BaseAvatarizationParameters, BaseAvatarizationWithTimeSeriesParameters
            ]
        ],
        Field(
            description="Avatarization parameters for each table. ",
            min_length=1,
            title="Table Parameters",
        ),
    ]
    seed: Annotated[
        Optional[int],
        Field(
            description="Seed used to generate the random numbers.If not provided, a random seed will be generated.",
            title="Seed",
        ),
    ] = None


class AvatarizationMultiTableResult(BaseModel):
    datasets: Annotated[List[AvatarizationResultPerDataset], Field(title="Datasets")]


class AvatarizationWithTimeSeriesJobCreate(BaseModel):
    parameters: AvatarizationWithTimeSeriesParameters


class PrivacyMetricsGeolocationJobCreate(BaseModel):
    parameters: PrivacyMetricsGeolocationParameters


class PrivacyMetricsMultiTableJobCreate(BaseModel):
    parameters: PrivacyMetricsMultiTableParameters


class AdviceJob(BaseModel):
    id: Annotated[UUID, Field(title="Id")]
    kind: JobKind
    created_at: Annotated[datetime, Field(title="Created At")]
    status: JobStatus
    error_message: Annotated[Optional[str], Field(title="Error Message")] = None
    traceback: Annotated[Optional[str], Field(title="Traceback")] = None
    result: Optional[Advice] = None
    parameters: AdviceParameters
    current_progress: Optional[JobProgress] = None


class AvatarizationMultiTableJob(BaseModel):
    id: Annotated[UUID, Field(title="Id")]
    kind: JobKind
    created_at: Annotated[datetime, Field(title="Created At")]
    status: JobStatus
    error_message: Annotated[Optional[str], Field(title="Error Message")] = None
    traceback: Annotated[Optional[str], Field(title="Traceback")] = None
    result: Optional[AvatarizationMultiTableResult] = None
    parameters: AvatarizationMultiTableParameters
    current_progress: Optional[JobProgress] = None


class AvatarizationMultiTableJobCreate(BaseModel):
    parameters: AvatarizationMultiTableParameters


@runtime_checkable
class Processor(Protocol):
    def preprocess(self, df: pd.DataFrame) -> pd.DataFrame: ...

    def postprocess(self, source: pd.DataFrame, dest: pd.DataFrame) -> pd.DataFrame: ...


class AvatarizationPipelineCreate(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    avatarization_job_create: AvatarizationJobCreate
    processors: List[Processor] = []
    df: pd.DataFrame


class AvatarizationPipelineResult(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    privacy_metrics: PrivacyMetrics
    signal_metrics: SignalMetrics
    post_processed_avatars: pd.DataFrame
    avatarization_job_id: UUID
    signal_job_id: UUID
    privacy_job_id: UUID



import pandas as pd
import pytest

from avatars.processors import GroupModalitiesProcessor


def test_postprocess_noop(many_dtypes_df: pd.DataFrame) -> None:
    """Check postprocess doing nothing with high parameter."""
    processor = GroupModalitiesProcessor(
        min_unique=3, global_threshold=1, new_category="other"
    )
    pd.testing.assert_frame_equal(
        many_dtypes_df, processor.postprocess(many_dtypes_df, many_dtypes_df)
    )


def test_group_modalities_simple(many_dtypes_df: pd.DataFrame) -> None:
    """Check preprocess group all value to modality as expected."""
    processor = GroupModalitiesProcessor(
        min_unique=2, global_threshold=1, new_category="other"
    )

    df_processed = processor.preprocess(many_dtypes_df)

    expected = many_dtypes_df.copy()
    expected["strings"] = "other"

    pd.testing.assert_frame_equal(expected, df_processed)


def test_group_preprocess_min_unique(categorical_df: pd.DataFrame) -> None:
    """Verify min_unique and threshold parameter work as expected."""
    df = categorical_df
    processor = GroupModalitiesProcessor(
        min_unique=2, global_threshold=1, new_category="other"
    )

    df_processed = processor.preprocess(df)

    expected = df.copy()
    expected["variable_1"] = ["other", "blue", "blue", "other"]

    pd.testing.assert_frame_equal(expected, df_processed)


def test_group_preprocess_new_category(categorical_df: pd.DataFrame) -> None:
    """Check the new category parameter works as expected."""
    df = categorical_df
    processor = GroupModalitiesProcessor(
        min_unique=2, global_threshold=1, new_category="andere"
    )

    df_processed = processor.preprocess(df)

    expected = df.copy()
    expected["variable_1"] = ["andere", "blue", "blue", "andere"]

    pd.testing.assert_frame_equal(expected, df_processed)


def test_group_preprocess_variable_thresholds(categorical_df: pd.DataFrame) -> None:
    """Verify the variable_thresholds parameter works as expected."""
    df = categorical_df
    processor = GroupModalitiesProcessor(
        variable_thresholds={"variable_1": 1, "variable_2": 2}
    )

    df_processed = processor.preprocess(df)

    expected = df.copy()
    expected["variable_1"] = ["other", "blue", "blue", "other"]
    expected["variable_2"] = ["other", "other", "other", "other"]

    pd.testing.assert_frame_equal(expected, df_processed)


def test_wrong_variable_thresholds_raises_error(categorical_df: pd.DataFrame) -> None:
    """Verify that an unknown variable in variable_thresholds raises an error."""
    processor = GroupModalitiesProcessor(variable_thresholds={"wrong_variable": 3})
    df = categorical_df
    with pytest.raises(
        ValueError, match="Expected valid variables in variable_thresholds*"
    ):
        processor.preprocess(df=df)


def test_with_global_threshold_without_min_unique() -> None:
    """Verify that providing only global_threshold raises an error."""
    with pytest.raises(
        ValueError, match="Expected both of (global_threshold, min_unique)*"
    ):
        GroupModalitiesProcessor(global_threshold=10)


def test_with_min_unique_without_global_threshold() -> None:
    """Verify that providing only min_unique raises an error."""
    with pytest.raises(
        ValueError, match="Expected both of (global_threshold, min_unique)*"
    ):
        GroupModalitiesProcessor(min_unique=10)


def test_variables_with_multiple_threshold_parameters() -> None:
    """Verify that providing global_threshold and variable_thresholds together raises an error."""
    with pytest.raises(
        ValueError, match="Expected variable_thresholds or (threshold, min_unique)*"
    ):
        GroupModalitiesProcessor(
            variable_thresholds={"wrong_variable": 3},
            global_threshold=10,
            min_unique=40,
        )



import numpy as np
import pandas as pd
import pytest

from avatars.processors import PerturbationProcessor


@pytest.mark.parametrize("perturbation_level", np.arange(0.1, 1, 0.1))
def test_does_perturb_continuous(perturbation_level: float) -> None:
    column = "a"
    source = pd.DataFrame([[0]], columns=[column])
    dest = pd.DataFrame([[1]], columns=[column])

    processor = PerturbationProcessor(perturbation_level={column: perturbation_level})
    expected = pd.DataFrame([[perturbation_level]], columns=[column])

    result = processor.postprocess(source, dest)
    pd.testing.assert_frame_equal(result, expected)


def test_does_perturb_categorical() -> None:
    column = "a"
    source = pd.DataFrame([1, 2, 3], columns=[column], dtype="object")
    dest = pd.DataFrame([10, 20, 30], columns=[column], dtype="object")

    processor = PerturbationProcessor(perturbation_level={column: 0.5}, seed=1)
    expected = pd.DataFrame([10, 20, 3], columns=[column], dtype="object")

    result = processor.postprocess(source, dest)
    pd.testing.assert_frame_equal(result, expected)


@pytest.mark.parametrize("perturbation_level", np.arange(0.1, 1, 0.1))
def test_categorical_proportions_are_respected(perturbation_level: float) -> None:
    """Verify that the perturbation level changes the proportion of avatars in the output."""
    shape = 100
    column = "a"

    source = pd.DataFrame(np.zeros(shape), columns=[column], dtype="object")
    dest = pd.DataFrame(np.ones(shape), columns=[column], dtype="object")

    processor = PerturbationProcessor(
        perturbation_level={column: perturbation_level}, seed=1
    )

    result = processor.postprocess(source, dest)

    # Counts the number of outputs that correspond to `dest` and make sure they are within
    # a certain tolerance.
    perturbed_counts = result.value_counts()[1.0]

    tolerance = 0.05
    upper_bound = shape * (perturbation_level + tolerance)
    lower_bound = shape * (perturbation_level - tolerance)

    assert lower_bound < perturbed_counts < upper_bound


def test_categorical_proportions_are_respected_edge_case() -> None:
    """Verify that a perturbation level of 1 and 0 keeps all and no perturbation."""
    shape = 10
    column = "a"

    source = pd.DataFrame(np.zeros(shape), columns=[column], dtype="object")
    dest = pd.DataFrame(np.ones(shape), columns=[column], dtype="object")

    # Perturbation level of one --> output should be the same as dest
    processor = PerturbationProcessor(perturbation_level={column: 1}, seed=1)
    result = processor.postprocess(source, dest)
    pd.testing.assert_frame_equal(dest, result)

    # Perturbation level of zero --> output should be the same as source
    processor = PerturbationProcessor(perturbation_level={column: 0}, seed=1)
    result = processor.postprocess(source, dest)
    pd.testing.assert_frame_equal(source, result)


def test_preprocess_noop(many_dtypes_df: pd.DataFrame) -> None:
    df = many_dtypes_df
    column = df.columns[0]
    processor = PerturbationProcessor(perturbation_level={column: 0.5})
    pd.testing.assert_frame_equal(df, processor.preprocess(df))


def test_missing_variable_raises_value_error(many_dtypes_df: pd.DataFrame) -> None:
    values = {"missing_variable": 1.0}
    processor = PerturbationProcessor(perturbation_level=values)
    with pytest.raises(ValueError, match="perturbation_level"):
        processor.postprocess(many_dtypes_df, many_dtypes_df)



from __future__ import annotations

from contextlib import contextmanager
from enum import Enum
from typing import Any, Generator, Optional, TypeVar

from toolz.dicttoolz import valfilter, valmap

T = TypeVar("T")
U = TypeVar("U")
R = TypeVar("R")


class ContentType(Enum):
    MULTIPART_FROM_DATA = "multipart/form-data"
    CSV = "text/csv"
    PDF = "application/pdf"
    JSON = "application/json"
    OCTET_STREAM = "application/octet-stream"
    ARROW_STREAM = "application/vnd.apache.arrow.stream"
    UNSUPPORTED = "unsupported"

    @classmethod
    def _missing_(cls, value: object) -> ContentType:
        return ContentType.UNSUPPORTED


def ensure_valid(what: Optional[T], label: str = "") -> T:
    if what is None:
        msg = f"{label} " if label else ""
        raise RuntimeError(f"Expected valid {msg} argument, got None instead")

    return what


@contextmanager
def validated(what: Optional[T], label: str = "") -> Generator[T, None, None]:
    val = ensure_valid(what, label)
    yield val


def pop_or(content: dict[str, Any], key: str, default: T) -> T:
    val: T = default

    if key in content:
        val = content.pop(key)

    return val


def remove_optionals(params: Optional[dict[str, Any]]) -> Optional[dict[str, Any]]:
    if params:
        # Remove params if they are set to None (allow handling of optionals)
        params = valfilter(lambda x: x is not None, params)
        params = valmap(lambda x: x.value if isinstance(x, Enum) else x, params)

    return params



from typing import Callable, Optional

import httpx

from avatars.client import ApiClient

RequestHandle = Callable[[httpx.Request], httpx.Response]


def mock_httpx_client(handler: Optional[RequestHandle] = None) -> httpx.Client:
    """Generate a HTTPX client with a MockTransport."""

    if handler is None:
        handler = lambda request: httpx.Response(200, json={})  # noqa: E731

    transport = httpx.MockTransport(handler)
    return httpx.Client(base_url="http://localhost:8000", transport=transport)


def api_client_factory(handler: Optional[RequestHandle] = None) -> ApiClient:
    """Generate an API client with a mock transport.

    The handler returns an empty 200 response by default.
    Consider overriding it with a custom handler for more complex tests.
    """
    http_client = mock_httpx_client(handler)
    return ApiClient(
        base_url="http://localhost:8000",
        http_client=http_client,
        verify_auth=False,
        should_verify_compatibility=False,
    )



# This file has been generated - DO NOT MODIFY
# API Version : 10.2.0


import logging
from io import BytesIO, StringIO
from typing import TYPE_CHECKING, Any, Dict, List, Optional, TypeVar, Union

from avatars.models import AdviceJob  # noqa: F401
from avatars.models import AdviceJobCreate  # noqa: F401
from avatars.models import AvatarizationJob  # noqa: F401
from avatars.models import AvatarizationJobCreate  # noqa: F401
from avatars.models import AvatarizationMultiTableJob  # noqa: F401
from avatars.models import AvatarizationMultiTableJobCreate  # noqa: F401
from avatars.models import AvatarizationWithTimeSeriesJob  # noqa: F401
from avatars.models import AvatarizationWithTimeSeriesJobCreate  # noqa: F401
from avatars.models import ClusterStats  # noqa: F401
from avatars.models import CompatibilityResponse  # noqa: F401
from avatars.models import Contributions  # noqa: F401
from avatars.models import CreateDataset  # noqa: F401
from avatars.models import CreateUser  # noqa: F401
from avatars.models import Dataset  # noqa: F401
from avatars.models import ExplainedVariance  # noqa: F401
from avatars.models import ForgottenPasswordRequest  # noqa: F401
from avatars.models import GenericJob  # noqa: F401
from avatars.models import Login  # noqa: F401
from avatars.models import LoginResponse  # noqa: F401
from avatars.models import PatchDataset  # noqa: F401
from avatars.models import PrivacyMetricsGeolocationJob  # noqa: F401
from avatars.models import PrivacyMetricsGeolocationJobCreate  # noqa: F401
from avatars.models import PrivacyMetricsJob  # noqa: F401
from avatars.models import PrivacyMetricsJobCreate  # noqa: F401
from avatars.models import PrivacyMetricsMultiTableJob  # noqa: F401
from avatars.models import PrivacyMetricsMultiTableJobCreate  # noqa: F401
from avatars.models import PrivacyMetricsWithTimeSeriesJob  # noqa: F401
from avatars.models import PrivacyMetricsWithTimeSeriesJobCreate  # noqa: F401
from avatars.models import Projections  # noqa: F401
from avatars.models import Report  # noqa: F401
from avatars.models import ReportCreate  # noqa: F401
from avatars.models import ReportFromDataCreate  # noqa: F401
from avatars.models import ReportGeolocationPrivacyCreate  # noqa: F401
from avatars.models import ResetPasswordRequest  # noqa: F401
from avatars.models import SignalMetricsJob  # noqa: F401
from avatars.models import SignalMetricsJobCreate  # noqa: F401
from avatars.models import SignalMetricsWithTimeSeriesJob  # noqa: F401
from avatars.models import SignalMetricsWithTimeSeriesJobCreate  # noqa: F401
from avatars.models import User  # noqa: F401
from avatars.models import FileType

if TYPE_CHECKING:
    from avatars.client import ApiClient


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
DEFAULT_RETRY_TIMEOUT = 60
DEFAULT_TIMEOUT = 60


T = TypeVar("T")


class Auth:
    def __init__(self, client: "ApiClient") -> None:
        self.client = client

    def login(
        self,
        request: Login,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> LoginResponse:
        """Login the user."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/login",  # noqa: F541
            "timeout": timeout,
            "form_data": request,
            "should_verify_auth": False,
        }

        return LoginResponse(**self.client.request(**kwargs))

    def refresh(
        self,
        token: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> LoginResponse:

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/refresh",  # noqa: F541
            "timeout": timeout,
            "params": dict(
                token=token,
            ),
        }

        return LoginResponse(**self.client.request(**kwargs))

    def forgotten_password(
        self,
        request: ForgottenPasswordRequest,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/login/forgotten_password",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
            "should_verify_auth": False,
        }

        return self.client.request(**kwargs)

    def reset_password(
        self,
        request: ResetPasswordRequest,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/login/reset_password",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
            "should_verify_auth": False,
        }

        return self.client.request(**kwargs)


class Compatibility:
    def __init__(self, client: "ApiClient") -> None:
        self.client = client

    def is_client_compatible(
        self,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> CompatibilityResponse:
        """Verify if the client is compatible with the API."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/check_client",  # noqa: F541
            "timeout": timeout,
            "should_verify_auth": False,
        }

        return CompatibilityResponse(**self.client.request(**kwargs))


class Datasets:
    def __init__(self, client: "ApiClient") -> None:
        self.client = client

    def create_dataset_from_stream(
        self,
        request: Union[StringIO, BytesIO],
        name: Optional[str] = None,
        filetype: Optional[FileType] = None,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Dataset:
        """Create a dataset by streaming chunks of the dataset."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/datasets/stream",  # noqa: F541
            "timeout": timeout,
            "params": dict(
                name=name,
                filetype=filetype,
            ),
            "file": request,
        }

        return Dataset(**self.client.request(**kwargs))

    def find_all_datasets_by_user(
        self,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> List[Dataset]:
        """List all datasets of the current_user."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/datasets",  # noqa: F541
            "timeout": timeout,
        }

        return [Dataset(**item) for item in self.client.request(**kwargs)]

    def create_dataset(
        self,
        request: Union[StringIO, BytesIO],
        name: Optional[str] = None,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Dataset:
        """Create a dataset from file upload."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/datasets",  # noqa: F541
            "timeout": timeout,
            "file": request,
            "form_data": dict(
                name=name,
            ),
        }

        return Dataset(**self.client.request(**kwargs))

    def get_dataset(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Dataset:
        """Get a dataset."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/datasets/{id}",  # noqa: F541
            "timeout": timeout,
        }

        return Dataset(**self.client.request(**kwargs))

    def patch_dataset(
        self,
        request: PatchDataset,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Dataset:
        """Modify a dataset."""

        kwargs: Dict[str, Any] = {
            "method": "patch",
            "url": f"/datasets/{id}",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return Dataset(**self.client.request(**kwargs))

    def analyze_dataset(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Dataset:
        """Start the analysis of a dataset."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/datasets/{id}/analyze",  # noqa: F541
            "timeout": timeout,
        }

        return Dataset(**self.client.request(**kwargs))

    def get_dataset_correlations(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:
        """Get a dataset's correlations."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/datasets/{id}/correlations",  # noqa: F541
            "timeout": timeout,
        }

        return self.client.request(**kwargs)

    def download_dataset_as_stream(
        self,
        id: str,
        filetype: Optional[FileType] = None,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:
        """Download a dataset by streaming chunks of it.

        Parameters
        ----------
        id
            The identifier of the dataset to download.
        filetype
            The filetype of the data you which to receive.

        Returns
        -------
            The dataset contents
        """

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/datasets/{id}/download/stream",  # noqa: F541
            "timeout": timeout,
            "params": dict(
                filetype=filetype,
            ),
            "should_stream": True,
        }

        return self.client.request(**kwargs)

    def download_dataset(
        self,
        id: str,
        filetype: Optional[FileType] = None,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:
        """Download a dataset.

        This is only advised for small datasets.
        Use /datasets/{id}/download/stream for larger datasets.
        """

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/datasets/{id}/download",  # noqa: F541
            "timeout": timeout,
            "params": dict(
                filetype=filetype,
            ),
        }

        return self.client.request(**kwargs)


class Health:
    def __init__(self, client: "ApiClient") -> None:
        self.client = client

    def get_root(
        self,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:
        """Verify server health."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/",  # noqa: F541
            "timeout": timeout,
        }

        return self.client.request(**kwargs)

    def get_health(
        self,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:
        """Verify server health."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/health",  # noqa: F541
            "timeout": timeout,
        }

        return self.client.request(**kwargs)

    def get_health_db(
        self,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:
        """Verify connection to the db health."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/health/db",  # noqa: F541
            "timeout": timeout,
        }

        return self.client.request(**kwargs)


class Jobs:
    def __init__(self, client: "ApiClient") -> None:
        self.client = client

    def find_all_jobs_by_user(
        self,
        nb_days: Optional[int] = None,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> List[GenericJob]:
        """Retrieve all jobs executed by the current user.

        Jobs are filtered by execution date, by default only the last 5 days are displayed,
        a parameter can be provided to go further back in time.
        """

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/jobs",  # noqa: F541
            "timeout": timeout,
            "params": dict(
                nb_days=nb_days,
            ),
        }

        return [GenericJob(**item) for item in self.client.request(**kwargs)]

    def create_full_avatarization_job(
        self,
        request: AvatarizationJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AvatarizationJob:
        """Create an avatarization job, then calculate metrics."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/jobs",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return AvatarizationJob(**self.client.request(**kwargs))

    def cancel_job(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> GenericJob:
        """Cancel any kind of job.

        If the job hadn't been started yet, revoke it.
        If the job is ongoing, gently kill it.
        If the job is done, do nothing.
        """

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/jobs/{id}/cancel",  # noqa: F541
            "timeout": timeout,
        }

        return GenericJob(**self.client.request(**kwargs))

    def create_avatarization_job(
        self,
        request: AvatarizationJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AvatarizationJob:
        """Create an avatarization job."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/jobs/avatarization",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return AvatarizationJob(**self.client.request(**kwargs))

    def create_avatarization_with_time_series_job(
        self,
        request: AvatarizationWithTimeSeriesJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AvatarizationWithTimeSeriesJob:
        """Create an avatarization with time series job."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/jobs/avatarization_with_time_series",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return AvatarizationWithTimeSeriesJob(**self.client.request(**kwargs))

    def create_avatarization_multi_table_job(
        self,
        request: AvatarizationMultiTableJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AvatarizationMultiTableJob:
        """Create an avatarization for relational data."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/jobs/avatarization_multi_table",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return AvatarizationMultiTableJob(**self.client.request(**kwargs))

    def create_signal_metrics_job(
        self,
        request: SignalMetricsJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> SignalMetricsJob:
        """Create a signal metrics job."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/jobs/metrics/signal",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return SignalMetricsJob(**self.client.request(**kwargs))

    def create_privacy_metrics_job(
        self,
        request: PrivacyMetricsJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsJob:
        """Create a privacy metrics job."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/jobs/metrics/privacy",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return PrivacyMetricsJob(**self.client.request(**kwargs))

    def create_privacy_metrics_time_series_job(
        self,
        request: PrivacyMetricsWithTimeSeriesJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsWithTimeSeriesJob:
        """Create a privacy metrics with time series job."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/jobs/metrics/privacy_time_series",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return PrivacyMetricsWithTimeSeriesJob(**self.client.request(**kwargs))

    def create_signal_metrics_time_series_job(
        self,
        request: SignalMetricsWithTimeSeriesJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> SignalMetricsWithTimeSeriesJob:
        """Create a signal metrics with time series job."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/jobs/metrics/signal_time_series",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return SignalMetricsWithTimeSeriesJob(**self.client.request(**kwargs))

    def create_privacy_metrics_multi_table_job(
        self,
        request: PrivacyMetricsMultiTableJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsMultiTableJob:
        """Create a privacy metrics job."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/jobs/metrics/privacy_multi_table",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return PrivacyMetricsMultiTableJob(**self.client.request(**kwargs))

    def create_privacy_metrics_geolocation_job(
        self,
        request: PrivacyMetricsGeolocationJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsGeolocationJob:
        """Create a geolocation privacy metrics job."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/jobs/metrics/privacy_geolocation",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return PrivacyMetricsGeolocationJob(**self.client.request(**kwargs))

    def get_privacy_metrics_geolocation_job(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsGeolocationJob:
        """Get a geolocation privacy metrics job."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/jobs/{id}/metrics/privacy_geolocation",  # noqa: F541
            "timeout": timeout,
        }

        return self.client.get_job(
            response_cls=PrivacyMetricsGeolocationJob,
            per_request_timeout=per_request_timeout,
            **kwargs,
        )

    def get_avatarization_job(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AvatarizationJob:
        """Get an avatarization job."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/jobs/avatarization/{id}",  # noqa: F541
            "timeout": timeout,
        }

        return self.client.get_job(
            response_cls=AvatarizationJob,
            per_request_timeout=per_request_timeout,
            **kwargs,
        )

    def get_avatarization_time_series_job(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AvatarizationWithTimeSeriesJob:
        """Get an avatarization time series job."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/jobs/avatarization_with_time_series/{id}",  # noqa: F541
            "timeout": timeout,
        }

        return self.client.get_job(
            response_cls=AvatarizationWithTimeSeriesJob,
            per_request_timeout=per_request_timeout,
            **kwargs,
        )

    def get_avatarization_multi_table_job(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AvatarizationMultiTableJob:
        """Get a multi table avatarization job."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/jobs/avatarization_multi_table/{id}",  # noqa: F541
            "timeout": timeout,
        }

        return self.client.get_job(
            response_cls=AvatarizationMultiTableJob,
            per_request_timeout=per_request_timeout,
            **kwargs,
        )

    def get_signal_metrics(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> SignalMetricsJob:
        """Get a signal metrics job."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/jobs/{id}/metrics/signal",  # noqa: F541
            "timeout": timeout,
        }

        return self.client.get_job(
            response_cls=SignalMetricsJob,
            per_request_timeout=per_request_timeout,
            **kwargs,
        )

    def get_signal_metrics_time_series_job(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> SignalMetricsWithTimeSeriesJob:
        """Get a signal metrics time series job."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/jobs/{id}/metrics/signal_time_series",  # noqa: F541
            "timeout": timeout,
        }

        return self.client.get_job(
            response_cls=SignalMetricsWithTimeSeriesJob,
            per_request_timeout=per_request_timeout,
            **kwargs,
        )

    def get_privacy_metrics(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsJob:
        """Get a privacy metrics job."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/jobs/{id}/metrics/privacy",  # noqa: F541
            "timeout": timeout,
        }

        return self.client.get_job(
            response_cls=PrivacyMetricsJob,
            per_request_timeout=per_request_timeout,
            **kwargs,
        )

    def get_privacy_metrics_time_series_job(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsWithTimeSeriesJob:
        """Get a privacy metrics time series job."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/jobs/{id}/metrics/privacy_time_series",  # noqa: F541
            "timeout": timeout,
        }

        return self.client.get_job(
            response_cls=PrivacyMetricsWithTimeSeriesJob,
            per_request_timeout=per_request_timeout,
            **kwargs,
        )

    def get_privacy_metrics_multi_table_job(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsMultiTableJob:
        """Get a privacy metrics multi table job."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/jobs/{id}/metrics/privacy_multi_table",  # noqa: F541
            "timeout": timeout,
        }

        return self.client.get_job(
            response_cls=PrivacyMetricsMultiTableJob,
            per_request_timeout=per_request_timeout,
            **kwargs,
        )

    def create_advice(
        self,
        request: AdviceJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AdviceJob:
        """Create advice on anonymization parameters."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/jobs/advice",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return AdviceJob(**self.client.request(**kwargs))

    def get_advice(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AdviceJob:
        """Get advice result."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/jobs/advice/{id}",  # noqa: F541
            "timeout": timeout,
        }

        return self.client.get_job(
            response_cls=AdviceJob, per_request_timeout=per_request_timeout, **kwargs
        )


class Metrics:
    def __init__(self, client: "ApiClient") -> None:
        self.client = client

    def get_job_projections(
        self,
        job_id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Projections:
        """Get the projections of records and avatars in 3D.

        See https://saiph.readthedocs.io/en/latest/ for more information.

        Arguments
        ---------
            job_id:
                avatarization or privacy job id used to fit the model
        """

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/projections/{job_id}",  # noqa: F541
            "timeout": timeout,
        }

        return Projections(**self.client.request(**kwargs))

    def get_variable_contributions(
        self,
        job_id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Contributions:
        """Get the contributions of the dataset variables within the fitted space.

        See https://saiph.readthedocs.io/en/latest for more information.

        Arguments
        ---------
            job_id:
                avatarization or privacy job id used to fit the model
        """

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/contributions",  # noqa: F541
            "timeout": timeout,
            "params": dict(
                job_id=job_id,
            ),
        }

        return Contributions(**self.client.request(**kwargs))

    def get_explained_variance(
        self,
        job_id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> ExplainedVariance:
        """Get the explained variance of records.

        See https://saiph.readthedocs.io/en/latest/ for more information.

        Arguments
        ---------
            job_id:
                avatarization or privacy job id used to fit the model
        """

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/variance/{job_id}",  # noqa: F541
            "timeout": timeout,
        }

        return ExplainedVariance(**self.client.request(**kwargs))


class Reports:
    def __init__(self, client: "ApiClient") -> None:
        self.client = client

    def create_report(
        self,
        request: ReportCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Report:
        """Create an anonymization report."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/reports",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return Report(**self.client.request(**kwargs))

    def get_report(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Report:

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/reports/jobs/{id}",  # noqa: F541
            "timeout": timeout,
        }

        return Report(**self.client.request(**kwargs))

    def download_report(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:
        """Download a report."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/reports/{id}/download",  # noqa: F541
            "timeout": timeout,
        }

        return self.client.request(**kwargs)

    def create_report_from_data(
        self,
        request: ReportFromDataCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Report:
        """Create an anonymization report without avatarization job."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/reports/from_data",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return Report(**self.client.request(**kwargs))

    def create_geolocation_privacy_report(
        self,
        request: ReportGeolocationPrivacyCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Report:
        """Create an anonymization report without avatarization job."""

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/reports/geolocation_privacy",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return Report(**self.client.request(**kwargs))


class Stats:
    def __init__(self, client: "ApiClient") -> None:
        self.client = client

    def get_cluster_stats(
        self,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> ClusterStats:
        """Get insights into the cluster's usage."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/stats/cluster",  # noqa: F541
            "timeout": timeout,
        }

        return ClusterStats(**self.client.request(**kwargs))


class Users:
    def __init__(self, client: "ApiClient") -> None:
        self.client = client

    def find_users(
        self,
        email: Optional[str] = None,
        username: Optional[str] = None,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> List[User]:
        """Get users, optionally filtering them by username or email.

        This endpoint is protected with rate limiting.
        """

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/users",  # noqa: F541
            "timeout": timeout,
            "params": dict(
                email=email,
                username=username,
            ),
        }

        return [User(**item) for item in self.client.request(**kwargs)]

    def create_user(
        self,
        request: CreateUser,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> User:
        """Create a user.

        This endpoint is protected with rate limiting.
        """

        kwargs: Dict[str, Any] = {
            "method": "post",
            "url": f"/users",  # noqa: F541
            "timeout": timeout,
            "json_data": request,
        }

        return User(**self.client.request(**kwargs))

    def get_me(
        self,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> User:
        """Get my own user."""

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/users/me",  # noqa: F541
            "timeout": timeout,
        }

        return User(**self.client.request(**kwargs))

    def get_user(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> User:
        """Get a user by id.

        This endpoint is protected with rate limiting.
        """

        kwargs: Dict[str, Any] = {
            "method": "get",
            "url": f"/users/{id}",  # noqa: F541
            "timeout": timeout,
        }

        return User(**self.client.request(**kwargs))



"""Original source: https://github.com/cgdeboer/iteround."""

from collections import OrderedDict
from typing import NamedTuple
from typing import OrderedDict as OrderedDictType
from typing import Tuple

import pytest

from avatars.lib.saferound import RoundingStrategy, saferound


class Setup(NamedTuple):
    in_list: list[float]
    in_dict: dict[str, float]
    in_odict: OrderedDictType[str, float]
    in_tuple: Tuple[float, float, float]
    neg_in_list: list[float]
    huge_in_list: list[float]


@pytest.fixture
def setup() -> Setup:
    in_list = [4.0001, 3.2345, 3.2321, 6.4523, 5.3453, 7.3422]
    in_dict = {
        "foo": 60.19012964572332,
        "bar": 15.428802458406679,
        "baz": 24.381067895870007,
    }
    in_odict = OrderedDict(in_dict)
    in_tuple = (60.1901296, 15.42880, 24.38106789)

    neg_in_list = [x * -1.0 for x in in_list]
    huge_in_list = [x * 100000.0 for x in in_list]
    setup = Setup(in_list, in_dict, in_odict, in_tuple, neg_in_list, huge_in_list)
    return setup


def test_basic_difference(setup: Setup) -> None:
    out = [4.0, 3.24, 3.23, 6.45, 5.35, 7.34]
    assert saferound(setup.in_list, 2) == out


def test_basic_largest(setup: Setup) -> None:
    out = [4.0, 3.23, 3.23, 6.45, 5.35, 7.35]
    assert saferound(setup.in_list, 2, RoundingStrategy.LARGEST) == out


def test_basic_smallest(setup: Setup) -> None:
    out = [4.0, 3.24, 3.23, 6.45, 5.35, 7.34]
    assert saferound(setup.in_list, 2, RoundingStrategy.SMALLEST) == out


def test_dict(setup: Setup) -> None:
    out = {"foo": 60.0, "bar": 16.0, "baz": 24.0}
    assert saferound(setup.in_dict, 0) == out


def test_odict(setup: Setup) -> None:
    out = OrderedDict({"foo": 60.0, "bar": 16.0, "baz": 24.0})
    assert saferound(setup.in_odict, 0) == out


def test_tuple(setup: Setup) -> None:
    out = (60.0, 16.0, 24.0)
    assert saferound(setup.in_tuple, 0) == out


def test_error_bad_float() -> None:
    bad = ["a", 3.2345, 3.2321, 6.4523, 5.3453, 7.3422]
    with pytest.raises(ValueError, match="all values in the iterable to be float"):
        saferound(bad, 2)


def test_error_bad_strategy(setup: Setup) -> None:
    with pytest.raises(ValueError, match="valid RoundingStrategy"):
        saferound(setup.in_list, 2, strategy="bad")  # type: ignore[arg-type]


def test_negative(setup: Setup) -> None:
    out = [-4.0, -3.24, -3.23, -6.45, -5.35, -7.34]
    assert saferound(setup.neg_in_list, 2) == out


def test_huge(setup: Setup) -> None:
    out = [400000.0, 324000.0, 323000.0, 645000.0, 535000.0, 734000.0]
    assert saferound(setup.huge_in_list, -3) == out


def test_overround(setup: Setup) -> None:
    out = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    assert saferound(setup.in_list, -3) == out


def test_over_with_sum(setup: Setup) -> None:
    out = [0.0, 0.0, 0.0, 10.0, 10.0, 10.0]
    assert saferound(setup.in_list, -1) == out


def test_topline(setup: Setup) -> None:
    out = [4.0, 3.0, 3.0, 7.0, 5.0, 7.0]
    topline = 29
    actual_out: list[float] = saferound(setup.in_list, 0, topline=topline)
    actual_topline = sum(actual_out)
    assert actual_out == out
    assert topline == actual_topline


def test_topline_sum(setup: Setup) -> None:
    assert sum(saferound(setup.in_list, 0, topline=28)) == 28
    assert sum(saferound(setup.in_list, 0, topline=29)) == 29
    assert sum(saferound(setup.in_list, 0, topline=30)) == 30


def test_empty_returns_empty() -> None:
    out: list[float] = []
    topline = 1
    actual_out = saferound([], 0, topline=topline)
    assert actual_out == out



import pandas as pd


class InterRecordRangeDifferenceProcessor:
    """Processor to express the values of two related variables relative to previous records.

    This processor can be used only on data where there are several records for each individual.
    By this transformation, variables such as `var_a` and `var_b` whose values are cumulative over
    successive records `t` in the following way:
    var_a_t <= var_b_t <= var_a_t+1 <= var_b_t+1 <= var_a_t+2 ...

    will be expressed as:
    - a variable containing the first value of `var_a`.
    - a variable containing the difference from the previous record (i.e. `var_a_t` - `var_b_t-1`)
    - a variable containing the range between the start and end variables
    (i.e. `var_b_t` - `var_a_t`)

    This processor can be used to express a quantity that varies at each event (variation defined
    by a start and an end) but also that varies across successive events.

    This processor is not suitable for data where any of the variables passed as args contain
    missing values.

    Keyword Arguments
    -----------------
        id_variable:
            variable indicating which individual each row belongs to
        target_start_variable:
            variable representing the start of the range to transform
        target_end_variable:
            variable representing the end of the range to transform
        sort_by_variable:
            variable used to sort records for each id
        new_first_variable:
            name of the variable to be created to contain the first value of the target variable
        new_range_variable:
            name of the variable to be created to contain the range value
        new_difference_variable:
            name of the variable to be created to contain the difference value
        keep_record_order:
            If set to `True`, the postprocess will decode values respecting the record order given
            by `id_variable` and `sort_by_variable` from the source dataframe. This can only be set
            to `True` if the indices are the same between the source and dest dataframes passed as
            arguments to `postprocess`.
    Examples
    --------
    >>> df = pd.DataFrame(
    ...     {
    ...         "id": [1, 2, 1, 1, 2, 2],
    ...         "start": [7, 14, 6, 12, 10, 23],
    ...         "end": [8, 18, 7, 15, 12, 24],
    ...     }
    ... )
    >>> processor = InterRecordRangeDifferenceProcessor(
    ...     id_variable="id",
    ...     target_start_variable='start',
    ...     target_end_variable='end',
    ...     sort_by_variable="start",
    ...     new_first_variable="first_value",
    ...     new_range_variable='range_value',
    ...     new_difference_variable="value_difference",
    ...     keep_record_order=True
    ... )
    >>> processor.preprocess(df)
       id  range_value  first_value  value_difference
    0   1            1            6               0.0
    1   2            4           10               2.0
    2   1            1            6               0.0
    3   1            3            6               4.0
    4   2            2           10               0.0
    5   2            1           10               5.0

    The postprocess allows you to transform some preprocessed data back into its original format
    >>> preprocessed_df = pd.DataFrame(
    ...     {
    ...         "id": [1, 2, 1, 1, 2, 2],
    ...         "range_value": [1, 4, 1, 3, 2, 1],
    ...         "first_value": [6, 10, 6, 6, 10, 10],
    ...         "value_difference": [0, 2, 0, 4, 0, 5],
    ...     }
    ... )

    >>> processor.postprocess(df, preprocessed_df)
       id  start   end
    0   1    7.0   8.0
    1   2   14.0  18.0
    2   1    6.0   7.0
    3   1   12.0  15.0
    4   2   10.0  12.0
    5   2   23.0  24.0

    The postprocess can also be used on data where the number of records per individual is
    different than the original one. In such cases, the processor should be instantiated with the
    `keep_record_order` argument set to its default value `False`.
    In the example below, there is an extra record with the id 2.

    >>> processor = InterRecordRangeDifferenceProcessor(
    ...     id_variable="id",
    ...     target_start_variable='start',
    ...     target_end_variable='end',
    ...     sort_by_variable="start",
    ...     new_first_variable="first_value",
    ...     new_range_variable='range_value',
    ...     new_difference_variable="value_difference",
    ...     keep_record_order=False
    ... )
    >>> preprocessed_df = pd.DataFrame(
    ...     {
    ...         "id": [1, 2, 1, 1, 2, 2, 2],
    ...         "range_value": [1, 4, 1, 3, 2, 1, 2],
    ...         "first_value": [6, 10, 6, 6, 10, 10, 10],
    ...         "value_difference": [0, 2, 0, 4, 0, 5, 1],
    ...     }
    ... )
    >>> processor.postprocess(df, preprocessed_df)
       id  start   end
    0   1    6.0   7.0
    1   2   12.0  16.0
    2   1    7.0   8.0
    3   1   12.0  15.0
    4   2   16.0  18.0
    5   2   23.0  24.0
    6   2   25.0  27.0
    """

    def __init__(
        self,
        *,
        id_variable: str,
        target_start_variable: str,
        target_end_variable: str,
        sort_by_variable: str,
        new_first_variable: str,
        new_range_variable: str,
        new_difference_variable: str,
        keep_record_order: bool = False,
    ):
        self.id_variable = id_variable
        self.target_start_variable = target_start_variable
        self.target_end_variable = target_end_variable
        self.sort_by_variable = sort_by_variable
        self.new_first_variable = new_first_variable
        self.new_range_variable = new_range_variable
        self.new_difference_variable = new_difference_variable
        self.keep_record_order = keep_record_order

    def preprocess(self, df: pd.DataFrame) -> pd.DataFrame:
        variables_to_check = [
            self.id_variable,
            self.target_start_variable,
            self.target_end_variable,
            self.sort_by_variable,
        ]
        if len(set(variables_to_check).difference(df.columns.values)) > 0:
            raise ValueError(
                "Expected valid variable names for `id_variable`, `target_start_variable`, "
                "`target_end_variable` and `sort_by_variable`, got "
                f"'{self.id_variable}', '{self.target_start_variable}', "
                f"'{self.target_end_variable}' and '{self.sort_by_variable}' instead"
            )

        if df[variables_to_check].isnull().values.any():
            raise ValueError(
                "Expected no missing values for `id_variable`, `target_start_variable`, "
                "`target_end_variable` and `sort_by_variable`, got columns with nulls instead"
            )

        df = df.copy()

        # data needs to be sorted
        df = df.sort_values([self.id_variable, self.sort_by_variable])

        # store range value (difference between end and start)
        df[self.new_range_variable] = (
            df[self.target_end_variable] - df[self.target_start_variable]
        )

        # store first value of target start variable
        df[self.new_first_variable] = df.groupby(self.id_variable)[
            self.target_start_variable
        ].transform("first")

        # add a temp column with the value of the previous line
        df["__end_tmp__"] = df.groupby(self.id_variable)[
            self.target_end_variable
        ].shift()

        # store the difference between current and previous value
        df[self.new_difference_variable] = (
            df[self.target_start_variable] - df["__end_tmp__"]
        )
        df[self.new_difference_variable] = df[self.new_difference_variable].fillna(0)

        # remove the tmp column and the original target variables
        df = df.drop(
            columns=[
                self.target_start_variable,
                self.target_end_variable,
                "__end_tmp__",
            ]
        )

        # sort by index to return processed data in the original order
        df = df.sort_index()

        return df

    def postprocess(self, source: pd.DataFrame, dest: pd.DataFrame) -> pd.DataFrame:
        variables_to_check = [
            self.id_variable,
            self.new_first_variable,
            self.new_range_variable,
            self.new_difference_variable,
        ]
        if len(set(variables_to_check).difference(dest.columns.values)) > 0:
            raise ValueError(
                "Expected valid variable names for `id_variable`, `new_first_variable`, "
                "`new_range_variable` and `new_difference_variable`, "
                f"got '{self.id_variable}', '{self.new_first_variable}', "
                f"'{self.new_range_variable}' and '{self.new_difference_variable}' instead"
            )

        if self.sort_by_variable not in source.columns.values:
            raise ValueError(
                f"Expected a valid `sort_by_variable`, got '{self.sort_by_variable}' instead"
            )

        if dest[variables_to_check].isnull().values.any():
            raise ValueError(
                "Expected no missing values for `id_variable`, "
                "`new_first_variable`, `new_range_variable` and `new_difference_variable`, "
                "got columns with nulls instead"
            )

        if source[self.sort_by_variable].isnull().values.any():
            raise ValueError(
                "Expected no missing values for `sort_by_variable` in source, "
                "got column with nulls instead"
            )

        if (
            self.keep_record_order
            and len(set(source.index).symmetric_difference(dest.index)) > 0
        ):
            raise ValueError(
                "Expected `keep_record_order` to be `True` only if source and dest "
                "have same indices, got source and dest with different indices"
            )

        df = dest.copy()

        # sort values in the same way as they were ordered in preprocess
        if self.keep_record_order:
            ordered_indices = source.sort_values(
                [self.id_variable, self.sort_by_variable]
            ).index
            df = df.loc[ordered_indices]

        # calculate for each row the sum of ranges of all previous records
        df["__range_tmp__"] = df.groupby(self.id_variable)[
            self.new_range_variable
        ].shift()
        df["__range_tmp__"] = df["__range_tmp__"].fillna(0)
        df["__totalizer_tmp__"] = df.groupby(self.id_variable)[
            "__range_tmp__"
        ].transform(pd.Series.cumsum) + df.groupby(self.id_variable)[
            self.new_difference_variable
        ].transform(
            pd.Series.cumsum
        )

        # calculate for each row the start value as the sum of the first value and the previously
        # calculated sum of ranges of all previous records (`__totalizer_tmp__`)
        df[self.target_start_variable] = (
            df[self.new_first_variable] + df["__totalizer_tmp__"]
        )

        # calculate end values from start value and range value
        df[self.target_end_variable] = (
            df[self.target_start_variable] + df[self.new_range_variable]
        )

        # remove all temp columns and columns added at preprocessing
        df = df.drop(
            columns=[
                "__range_tmp__",
                "__totalizer_tmp__",
                self.new_difference_variable,
                self.new_range_variable,
                self.new_first_variable,
            ]
        )

        # sort by index to return data in the original order
        df = df.sort_index()
        return df



import numpy as np
import pandas as pd
import pandas.testing as pd_testing
import pytest

from avatars.processors import ToCategoricalProcessor


@pytest.fixture
def original() -> pd.DataFrame:
    df = pd.DataFrame(
        {
            "variable_1": [1, 2, 3, np.nan],
            "variable_2": [1, 2, 2, np.nan],
        }
    )
    return df


def test_preprocess_to_categorical_keep_continuous(original: pd.DataFrame) -> None:
    """Check preprocess with keep continuous."""
    expected = original.copy()
    expected["variable_2__cont"] = expected["variable_2"]
    expected = expected.astype({"variable_2": "object"})

    processor = ToCategoricalProcessor(to_categorical_threshold=2, keep_continuous=True)
    df = processor.preprocess(df=original)

    pd_testing.assert_frame_equal(expected, df)


def test_preprocess_to_categorical(original: pd.DataFrame) -> None:
    """Check preprocess without keep continuous."""
    expected = original.copy()
    expected = expected.astype({"variable_2": "object"})

    processor = ToCategoricalProcessor(to_categorical_threshold=2)
    processed = processor.preprocess(df=original)

    pd_testing.assert_frame_equal(expected, processed)


def test_processor_to_categorical(original: pd.DataFrame) -> None:
    """Verify that pre- and post-process, with keep_continuous=False, gives original."""
    expected = original.copy()
    processor = ToCategoricalProcessor(
        to_categorical_threshold=2, keep_continuous=False
    )
    processed_df = processor.preprocess(df=original)
    df = processor.postprocess(source=original, dest=processed_df)

    pd_testing.assert_frame_equal(df, expected)


def test_processor_to_categorical_keep_continuous(original: pd.DataFrame) -> None:
    """Verify that pre- and post-process, with keep_continuous=True, gives original."""
    expected = original.copy()
    processor = ToCategoricalProcessor(to_categorical_threshold=2, keep_continuous=True)
    processed_df = processor.preprocess(df=original)
    df = processor.postprocess(source=original, dest=processed_df)
    pd_testing.assert_frame_equal(df, expected)


def test_postprocessed_with_category(original: pd.DataFrame) -> None:
    """Check post processor with a transformed variable changed the categorical variable.

    Transformed data frame could be obtained with another processor such as GroupModalities().
    """
    transformed = pd.DataFrame(
        {
            "variable_1": [1, 2, 3, np.nan],
            "variable_2": ["other", "other", "other", np.nan],
            "variable_2__cont": [1, 2, 2, np.nan],
        }
    )
    processor = ToCategoricalProcessor(
        to_categorical_threshold=2, keep_continuous=True, category="other"
    )
    _ = processor.preprocess(
        df=original
    )  # Needed to to add self.continuous_variables to the model
    df = processor.postprocess(source=original, dest=transformed)
    pd_testing.assert_series_equal(df["variable_2"], original["variable_2"])
